                              (* BLOCKIN.PAS *)

{$R+}    (* Range checking on/off *)
{$I+}    (* I/O checking *)
{$D+}    (* debugging on/off; work with Options - Debugger - Integrated *)

program BlockIn;

(* Gibbs sampling estimation for multi-block model;
   programmed by Tom A.B. Snijders, University of Groningen.
   Refers to paper "Estimation and prediction for stochastic blockstructures"
   by Krzysztof Nowicki and Tom Snijders.

   This program BlockIn is a DOS version for the user interface,
   and makes the basic information file which is used
   by the computation part of the program. *)


Uses
   Blib, B_cons, Crt;
(* Blib is a general procedure library.
   B_cons contains constants. *)

type

BlockInMake = object

proname:         string;                               (* Project name *)

comin, comax:            integer;
missing:                 integer; (* missing number indicator *)
nrun0,nrun1:             longint;
theta00:    real;
(* theta and eta are parameters;
   theta0 and eta0 are their prior values.
   theta0 = T_k in the paper,
   and defined as theta0 = 1 + theta00 * c (where c = number of colors).
   Note that theta0 is identical over all coordinates. *)
identified:               boolean;
(* identified indicates whether the model is identified
   by prior probabilities for x *)
multiple, optimize, overdisx, overdiseta :
                        boolean;
(* multiple indicates more than one Gibbs sequence per model;
   optimize indicates optimization of starting values;
   overdisx and overdiseta indicate overdispersion of x and of eta
   in the first half of the pre-convergence iterations. *)
multi:         integer;
(* multi = total number of Gibbs sequences per model
           if option multiple is chosen *)
appendout: boolean;

procedure initialise;
procedure ask;

end; (* object BlockInMake *)

var

inf: text; (* basic information file *)
BIM: BlockInMake;


procedure BlockInMake.initialise;
var infn: string;
  sd: longint;
begin
(* Warning:
   I put this in earlier, but I think now it is superfluous.
   if (cmax > ccmax) then begin
    writeln('Warning: constant ccmax in BLIB should not be smaller ');
    writeln('than constant cmax in B_CONS.');
    writeln('Recompile B_CONS with smaller constant cmax.');
    waitkey;
    halt;
   end;

   *)

  writeln('Give the name of the project (root name).');
  write('Project name : ');
  readln(proname);
  infn := proname+'.in';
  assign(inf,infn);
  rewrite(inf);
end; (* initialise *)

procedure BlockInMake.Ask;
var i,cv,dum: integer;
  descriptions:  SSarray;
  choices: BBarray;
  dfn0,dfn,pfn: string;
  mode: char;
  v: array[1..cmax] of integer;
begin
  clrscr;
  writeln('The input data file must start with a title line');
  writeln('and then contain a square adjacency matrix');
  writeln('composed of integers between ',-cc,' and ',cc,',');
  writeln('separated by blanks.');
  writeln;
  dfn0 := proname+'.dat';
  writeln('Give name for input data file (default ',dfn0,'):');
  readln(dfn);
  if (dfn = '') then dfn := dfn0;
  writeln('The model specification is given interactively.');
  writeln('Symbols, sections, and formula numbers refer to the manuscript');
  writeln('by Nowicki and Snijders, which describes the estimation procedure.');
  writeln('For symbols, underscore _ denotes a subscript,');
  writeln('e.g., M_0 means M with subscript 0.');
  writeln;
  writeln('Give the missing value indicator (between ',-cc,' and ',cc,'): ');
  repeat
    write('(between ',-cc,' and ',cc,') : ');
    readln(missing);
   until ((missing >= -cc) and (missing <= cc));
  writeln;

  writeln('Give the smallest and largest number (c) of latent classes.');
  write('Smallest number of latent classes ');
  repeat
    write('(between 1 and ',cmax,') : ');
    readln(comin);
   until (comin in [1..cmax]);
  writeln;
  write('Largest number of latent classes (= colors) ');
  repeat
    write('(between ',comin,' and ',cmax,') : ');
    readln(comax);
   until (comax in [comin..cmax]);
  writeln;
  write('Give the default number of runs (2M_0) before convergence: ');
  repeat
    readln(nrun0);
    if (nrun0 < 100) then write('(more than 99) : ');
   until (nrun0 >= 100);
  writeln;
  write('Give the default number of runs after convergence: ');
  repeat
    readln(nrun1);
    if (nrun1 < 100) then write('(more than 99) : ');
   until (nrun1 >= 100);

  writeln; writeln;
  writeln('Do you wish to identify the latent classes ');
  write('by prior probabilities for the colors (see Section 6.1) ? ');
  identified := AnswerYes;
  writeln;
  if (identified and (comin = comax)) then begin
    writeln('You may identify the colors in two ways:');
    writeln('1. The easy way: give vertex numbers identifying the colors.');
    writeln('2. The more complicated way: give prior probabilities in a file.');
    write('Choose the mode (1 or 2):');
    repeat readln(mode);
     if not (mode in ['1','2']) then write('between 1 and 2: ');
     until (mode in ['1','2']);
   end;
  if identified and (comax > comin) then mode := '1';
  if not identified then mode := '0';

  writeln; writeln;
  if (mode = '1') then begin
    for i:=1 to cmax do v[i] := 0;
    writeln('Give the vertices identifying the colors ');
    writeln('(number of identifying vertices must be between ',
            comin-1,' and ',comax,').');
    writeln('Use the vertex numbers occurring in the data set.');
    writeln('For each Gibbs sequence with c colors,');
    writeln('the first c mentioned vertices will be used.');
    writeln('Terminate each vertex number with the <Enter> key,');
    writeln('and terminate the entire sequence (if you give less than',
            comax,' vertices)');
    writeln('by the number 0.');
    writeln;
    cv := 0;
    repeat
      cv := cv+1;
      write('The ');
      case cv of
        1: write('1st ');
        2: write('2nd ');
        3: write('3d  ');
        else write(cv,'th ');
       end; (* case *)
      write('identifying vertex number:  ');
      readln(v[cv]);
     until ((cv >= comax) or (v[cv] = 0) or (cv >= cmax));
    end; (* if mode ' '1' *)

  if (mode = '2') then begin
    write('Give the name of the data file with the prior probabilities :');
    readln(pfn);
    writeln('Note that this file must contain a rectangular matrix of probabilities,');
    writeln('with ',comin,' columns (for the colors)');
    writeln('and as many rows as there are identifying vertices;');
    writeln('each row must contain first the vertex number, ');
    writeln('then the prior probabilities that the vertex has each of the colors;');
    writeln('these probabilities must be nonnegative and sum to 1;');
    writeln('all numbers in this file must be separated by blanks.');
   end; (* if *)

  writeln; writeln;
  writeln('The prior probabilities for the color probabilities theta');
  writeln('can be more or less concentrated near their average value of 1/c.');
  writeln('You can now give the concentration parameter.');
  writeln('A value of 0 will lead to a uniformly distributed prior for theta,');
  writeln('a value larger than 100 will imply a highly concentrated');
  writeln('prior for theta.');
  writeln('(The parameter T in Section 5 is then defined as 1 + c*con,');
  writeln(' where c is the number of colors and con is this concentration parameter.)');
  write('Give the value for the concentration parameter (suggestion: 100): ');
  repeat
    readln(theta00);
    if (theta00 < 0.0) then begin
      writeln('The concentration parameter must be 0 or positive.');
      write('Give the value for the concentration parameter: ');
     end;
   until (theta00 >= 0.0);

  writeln; writeln;
  writeln;
  writeln('You can choose a number of options that will be in effect');
  writeln('for each model estimated according to the given specifications.');
  writeln('See Section 5.2 for their motivation.');
  writeln('(If you choose for more than one Gibbs sequence per model,');
  writeln(' the search for good starting colorings will be automatic.)');
  writeln;

  writeln('Specification of the number of Gibbs sequences per model.');
  writeln('The answer 0 means that you will only carry out data inspection');
  writeln('and an initial search of (good or random) starting colorings.');
  writeln;
  write('Give the number of Gibbs sequences per model : ');
    repeat
      readln(multi);
      if (multi < 0) then write('at least 0 : ');
     until (multi >= 0);
  multiple := (multi >= 2);

  descriptions[1] := 'Search for good starting colorings                 ';
  descriptions[2] := 'Overdispersed colorings in pre-convergence iterations    ';
  descriptions[3] := 'Overdispersed probabilities in pre-convergence iterations';
(* Remark: the following true/false choices give the defaults,
   which can be superseded by the answer to the OptionChoice procedure. *)
  choices[1] := true;
  choices[2] := true;
  choices[3] := false;
  Optionchoice(3, 0, descriptions, choices);
  optimize   := choices[1];
  overdisx   := choices[2];
  overdiseta := choices[3];

  writeln('Specification of whether the output file must be rewritten or appended.');
  writeln;
  writeln('Indicate by 0 or 1 whether you want to append the output');
  writeln('to the existing output file (if it does exist) : ');
    repeat
      readln(dum);
     until ((dum = 0) or (dum = 1));
  appendout := (dum = 1);
(* write basic information file *)
  writeln(inf,dfn,'   (data file)');
  writeln(inf,comin:2,' ',comax:2,'      (minimum and maximum number of colors)' );
  writeln(inf,nrun0:6,'     (number of runs before convergence is assumed)');
  writeln(inf,nrun1:6,'     (number of runs after convergence is assumed)');
  writeln(inf,multi:2, '         (number of Gibbs sequences per color)');
  writeln(inf,' ',mode,'         (identification mode)');
  writeln(inf,missing:2,'         (missing number indicator)');
  writeln(inf,theta00:8:2,'   (concentration parameter)');
  if optimize then writeln(inf,
  ' 1         (search for good starting colorings)')
    else writeln(inf,' 0         (random starting colorings)');
  if overdisx then writeln(inf,
  ' 1         (overdispersed colorings in pre-convergence iterations)')
    else writeln(inf,' 0         (no overdispersed colorings in pre-convergence iterations)');
  if overdiseta then writeln(inf,
  ' 1         (overdispersed probabilities in pre-convergence iterations)')
    else writeln(inf,' 0         (no overdispersed probabilities in pre-convergence iterations)');
  case mode of
    '1': begin
           i := 0;
           repeat begin
             i := i+1;
             write(inf,v[i]:2,' ');
            end
           until ((i >= comax) or (i >= cmax) or (v[i] = 0));
          writeln(inf,' 0 ');
         end;
     '2': writeln(inf,pfn);
   end; (* case *)
   if appendout then write(inf,'1') else write(inf,'0');
   writeln(inf,'     (0 = rewrite, 1 = append output file)');
  close(inf);
end; (* ask *)


begin (* main *)

with BIM do begin
  initialise;
  ask;
 end;

end.

