                             (* MULBLO.PAS *)

unit MulBlo;

(* Gibbs sampling estimation for multi-block model;
   programmed by Tom A.B. Snijders, University of Groningen.
   Refers to paper "Estimation and prediction for stochastic blockstructures"
   by Krzysztof Nowicki and Tom Snijders.
   Most symbols in the program are identical to those in the paper.
   Alpha = set of tie values of relations
   AA = script A = alphabet of pairwise relations.
   The unit MULBLO contains the computational part .
    *)


interface

Uses
  WBlib, RanGen, B_cons, BlockMo, BlockI, Forms, Dialogs, SysUtils;
(* RanGen is a library of random variable generators;
   WBlib is a general procedure library;
   BlockMo contains object BlockModel.*)


type MultiGibbs = object(Blockmodel)

theta:                array[1..cmax] of real;
theta0,w0:    real;
eta,eta0:             array[1..rmax, 1..cmax, 1..cmax] of real;
(* theta and eta are parameters;
   theta0 and eta0 are their prior values.
   theta0 = T_k in the paper,
   and defined as theta0 = 1 + theta00 * c (where c = number of colors).
   Note that theta0 is identical over all coordinates.
   w0 is a factor by which the Dirichlet parameters for drawing eta
   are multiplied during the initial iterations. *)
x:                     coloring;
x0:                    array of array of integer;
(* x is the current coloring; x0 gives the initial colorings for the
   optimal starting points *)
xpair:                 array of array of real;
thetax, predm:         array of real;
theta2:                array[1..cmax] of real;
(* thetax gives posterior distribution of theta[x[i]] for non-identified,
    of theta[i] for identified models;
    theta2 used for the posterior standard deviations
    only in the identified case;
    predm gives the posterior predictive distribution of m[x[i]],
    the number of vertices with the same color as i.*)

etax, etax2:  array of array of array of real;
(* etax keeps the sums of eta[x[i],x[j]], etax2 their sums of squares *)
(* The following arrays are for identified models : *)
classprob:             array of array of real;  (* cf. xpair *)
etac, etac2:  array of array of array of real;
                       (* cf. etax and etax2 *)
entropyx,entropyxm,informationy,informationym:
                       real;
nstat,nentropy,nentropym:
                       longint;
EntrFrequency, EntrShowFrequency, EntrWriteFrequency,
ShowXEFrequency        : integer;
(* Note concerning the calculation of entropies.
   The entropies have been bracketed away, since they
   were canceled in the JASA paper.
   Once in every EntrFrequency runs, the entropy in x is calculated in
   Drawx and the information in y is calculated in Draweta.
   At the same frequency, calcstats is called.
   The partial sums are held in Entropyx and Informationy, the number
   of terms in Nentropy.
   As soon as Nentropy >= EntrShowFrequency,
   procedure ShowXandEnt is called in Gibbsing:
   the averages are calculated and displayed on the screen;
   - once every EntrWriteFrequency of these times,
   they are also displayed in the output file; -
   partial sums and numbers of terms of these displayed averages
   are stored in Entropyxm, Informationym, Nentropym;
   in Gibbsresults, the averages are calculated
   and displayed in the output file.*)


Converged:        boolean;
(* converged indicates whether it is assumed that the process has converged;
   calcsd indicates whether standard errors are to be computed;
   identified indicates whether the model is identified
   by prior probabilities for x;
   both indicates that first a non-identified model, and then
   an identified model is to be run;
   starting indicates that the program is in the starting phase. *)
width, multinum:         integer;
(*  multinum = current ordinal number of the Gibbs sequence for option multiple.
   width = number of spaces in primary output file
                         for rounded probabilities *)

procedure SetMulArrays(var outf: textfile);
procedure Gibbs(var outf, outf1: textfile; outfn, outfn1: string);
procedure PriorPriors;
procedure Gibbsing(var outf, outf1: textfile; outfn, outfn1: string);
procedure InitGibbs(var outf, outf1: textfile; outfn, outfn1: string);
procedure GibbsResults(var outf, outf1: textfile; outfn, outfn1: string;
                       naftercon: integer);
procedure FinishGibbs(var outf, outf1: textfile);
procedure stopsthis(var outf, outf1: textfile);


private
m,mm:       array[1..cmax] of integer;
(* mm[i] is number of vertices of color i;
   m[i] is the same, excluding vertices for which prior[i] *)
d:          array of array of array of integer;
e:          array[1..rmax,1..cmax,1..cmax] of longword;
procedure randomstart;
procedure optimalstart(var outf, outf1: textfile; outfn, outfn1: string);
procedure calccounts;
procedure calcstats;
procedure drawtheta;
procedure draweta(CalcEntropy: Boolean);
procedure drawx(CalcEntropy: boolean);
procedure onestep(CalcEntropy: boolean);
end; (* object MultiGibbs *)

//var lf: textfile;

implementation

uses blocks_1;

var ntiny: longword;

procedure MultiGibbs.SetMulArrays(var outf: textfile);
begin
  nforform := n; (* used in blocks_1 *)
  SetLength(etax,  r01max+1,n+1,n+1);
  SetLength(etax2, r01max+1,n+1,n+1);
  SetLength(x, n+1);
  SetLength(x0, mulmax+1,n+1);
  SetLength(xpair,n+1,n+1);
  SetLength(thetax,n+1);
  SetLength(predm,n+1);
  SetLength(classprob,n+1,cmax+1);
  SetLength(d,rmax+1,n+1,cmax+1);
  SetLength(etac,r01+1,cmax+1,cmax+1);
  SetLength(etac2,r01+1,cmax+1,cmax+1);
end; (* SetMulArrays *)


procedure MultiGibbs.PriorPriors;
var i,j,k: integer;
begin
(* Uniform prior distribution : *)
  theta0 := 1.0;
  w0 := 1.0;
  (* these values for theta0 and w0 are completely unimportant,
     and superseded in Gibbsing *)
  for i:=c downto 1 do begin
    for j:=i-1 downto 1 do begin
      for k:=r downto 1 do begin
        eta0[k,j,i] := 1.0;
        eta0[ppi[k],i,j] := eta0[k,j,i];
       end; (* k *)
     end; (* j *)
    for k:= r01 downto 1 do eta0[k,i,i] := 1.0;
    for k:= r01+1 to r do eta0[k,i,i] := eta0[ppi[k],i,i];
   end; (* i *)
   if (not IdentificationByFile) then begin
     for i:=1 to n do begin
       prior[i] := false;
       for j:=1 to cmax do thetai[i,j] := 0.0;
      end; (* for i *)
    end; (* if *)
end; (*  MultiGibbs.PriorPriors *)

procedure MultiGibbs.Gibbs(var outf, outf1: textfile; outfn, outfn1: string);
var col: integer;
    s: string;
begin
//  assignfile(lf,proname+'.log');
//  rewrite(lf);
//  writeln(lf,'Log file ',proname,'.log');
  Heading(lf,1,proname);
  for col:=comin to comax do if (not escape) then begin
    starting := true;
    c := col;
    s := st(c)+' colors.';
    Heading(outf,1,s);
    Heading(outf1,1,s);
    Heading(lf,2,s);
    if (col > comin) then
      begin
      ranseed := takeseed(3);
      writeln(outf,'Current random number generator seed is ',ranseed,'.');
      init3(ranseed);
      end;
    multinum := 0;
    if both then identified := false;
    PriorPriors;
    if optimize then OptimalStart(outf1,outf1,outfn,outfn1);
    (* note that both files are indicated as outf1,
       so all output of optimalstart goes to outf1. *)

    starting := false;
     if ((multi >= 1) and (not escape)) then begin
      repeat
        inc(multinum);
        s := 'Gibbs sequence number '+st(multinum)+'.';
        Heading(outf,2,s);
        Heading(outf1,2,s);
        Gibbsing(outf, outf1,outfn,outfn1);
        if (both  and (not escape)) then begin
          identified := true;
          s := 'Gibbs sequence number '+st(multinum)+' with identification.';
          Heading(outf,2,s);
          Heading(outf1,2,s);
          Gibbsing(outf, outf1,outfn,outfn1);
          identified := false;
         end; (* if both *)
       until ((multinum >= multi) or Escape);
     end (* if multiple >= 1 *)
     else writeln(outf,'No Gibbs sequences were requested.');
   end; (* for col *)
  Finalize(etax);
  Finalize(etax2);
  Finalize(x0);
  Finalize(xpair);
  Finalize(thetax);
  Finalize(predm);
  Finalize(classprob);
  Finalize(d);
  Finalize(y);
  Finalize(data);
  Finalize(prior);
  Finalize(thetai);
  Finalize(etac);
  Finalize(etac2);
//  closefile(lf);
end; (* Gibbs *)

procedure MultiGibbs.onestep(calcentropy: boolean);
begin
  drawtheta;
  draweta(CalcEntropy);
  drawx(calcentropy);
end; (* onestep *)

procedure MultiGibbs.gibbsing(var outf, outf1: textfile; outfn, outfn1: string);
var i, naftercon: longint;
  j,k: integer;
  calcentropy, Exist0: boolean;
  entropyx0, informationy0: real;
label Convergence;
label Startlabel;
procedure UpdateShowEnt;
var k: integer;
  onesequence, onecolor, done: integer;
begin
(* updates *)
  entropyx := entropyx / (n*nentropy);
  informationy := informationy / (ndat*nentropy);
  inc(nentropym);
  entropyxm := entropyxm + entropyx;
  informationym := informationym + informationy;

(* calculate relative number of iterations behind us *)
  onesequence := nrun0+nrun1;
  if both then onesequence := 2*onesequence;
  onecolor := multi*onesequence;
  done := (c-comin)*onecolor + (multinum-1)*onesequence + i;
  if converged then done := done + nrun0;
  if (both and identified) then done := done + nrun0 + nrun1;
  if ((nentropym mod ShowXEFrequency) = 0) then
    ShowXE(i,nforform,x,done);
  if ((nentropym mod EntrWriteFrequency) = 0) then begin
(*    write(outf,i:6,entropyx:8:4,informationy:8:4,'   '); *)
    write(outf,i:6,informationy:8:4,'   ');
    for k:=1 to n do write(outf,x[k]); writeln(outf);
   end;
  entropyx0 := entropyx;
  informationy0 := informationy;
  Exist0 := true;
  nentropy := 0;
  entropyx := 0.0;
  informationy := 0.0;
end; (* UpdateShowEnt *)
begin (* gibbsing *)
(* general initialisation *)
  Initgibbs(outf, outf1, outfn, outfn1);
  PriorPriors;
  if identified then Identifyvertices(outf);
  if optimize then
    for i:=1 to n do x[i] := x0[multinum,i]
   else RandomStart;
  calccounts;
  Exist0 := false;
  calcsd := true;

(* initialisation for pre-convergence runs *)
startlabel:
  i := 0;
  Nentropy := 0;
  entropyx := 0.0;
  informationy := 0.0;
  converged := false;
  EntrWriteFrequency := nrun0 div (EntrFrequency * EntrShowFrequency * 50);
  if (EntrWriteFrequency < 1) then EntrWriteFrequency := 1;
  writescreen(c,comin,comax,nrun0,identified,multiple,converged,multinum,multi);
  Heading(outf,3,'Iterations: values of some random variables');
  writeln(outf,'Before convergence is assumed');
  writeln(outf,'iter., Info. y, last x: ');
(* pre-convergence runs: *)
  repeat
    Application.ProcessMessages;
    Inc(i);
(* Overdispersed distributions in the first 0.8*nrun0 runs *)
    if (5 * i < 4*nrun0) then begin
(*      if overdisx then theta0 := (2*i/nrun0) + 4*n*(1.0 - (2*i/nrun0)) *)
   if overdisx then
      theta0 := (1 + theta00*c)*((1.25*i/nrun0) + 100*(1.0 - (1.25*i/nrun0)))
       else theta0 := 1 + theta00*c;
      if overdiseta then w0 := (2*i/nrun0)*((n-1)/n) + 1/n
       else w0 := 1.0;
     end else begin
      theta0 := 1 + theta00*c;
      w0 := 1.0;
     end;
    calcentropy := ((i mod EntrFrequency) = 0);
    onestep(calcentropy);
    if (calcentropy and (nentropy >= EntrShowFrequency)) then begin
     UpdateShowEnt;
    end;
    if Restart then goto StartLabel;
   until ((i >= nrun0 ) or Escape);

 if (not escape) then begin

  converged := true;
  writeln(outf);
  writeln(outf,i,' runs before convergence was assumed.');
  if (ntiny > i) then begin
    writeln(outf1,'There may be problems with numerical accuracy.');
    writeln(outf1,'In ',ntiny,' cases, truncation of tiny numbers was necessary.');
   end;


  naftercon := 0;
Convergence:
(* initialisation of after-convergence runs *)
  nstat := 0;
  ntiny := 0;
  Nentropy := 0;
  entropyx := 0.0;
  informationy := 0.0;
  nentropym := 0;
  entropyxm := 0.0;
  informationym := 0.0;
  writescreen(c,comin,comax,nrun1,identified,multiple,converged,multinum,multi);
  for i:=1 to cmax do theta2[i] := 0.0;
  for i:=1 to n do begin
    thetax[i] := 0.0;
    predm[i] := 0.0;
    for j:=1 to n do begin
      xpair[i,j] := 0.0;
      if not identified then
        for k:=r01 downto 1 do begin
          etax[k,i,j] := 0.0;
          if calcsd then etax2[k,i,j] := 0;
         end; (* k *)
     end; (* j *)
    for j:=1 to cmax do classprob[i,j] := 0.0;
   end; (* i *)
  if identified then
   for k:=r01 downto 1 do for i:=1 to cmax do for j:=1 to cmax do begin
    etac[k,i,j] := 0.0;
    if calcsd then etac2[k,i,j] := 0.0;
   end; (* for k,i,j *)

  i := 0;
  theta0 := 1 + theta00*c;
  w0 := 1.0;
  EntrWriteFrequency := nrun1 div (EntrFrequency * EntrShowFrequency * 50);
  if (EntrWriteFrequency < 1) then EntrWriteFrequency := 1;
  writeln(outf);
  writeln(outf,'After convergence is assumed');
  writeln(outf,'iteration i, Information y, last x: ');
(* After-convergence runs : *)
  repeat
    if Restart then begin
      writeln(outf,'Restart of post-convergence runs after ',i,' iterations.');
      naftercon := naftercon + i;
      restart1:=false;
      goto Convergence;
     end; (* if *)
    Inc(i);
    calcentropy := ((i mod EntrFrequency) = 0);
    onestep(calcentropy);

    if (calcentropy and (nentropy >= EntrShowFrequency)) then begin
     UpdateShowEnt;
    end;
    if calcentropy then calcstats;
   until ((i >= nrun1 ) or Escape);
//  stop1:=false; //pb

  naftercon := naftercon + i;
  if escape then writeln(outf,
    'The user asked for an early stop of this Gibbs sequence.');
  writeln(outf,'After assuming convergence, ',
               naftercon,' runs for estimating posterior distribution.');
  writeln(outf);
  Heading(outf,3,'Results');
  Heading(outf1,3,'Results');

  if (nstat >= 10) then GibbsResults(outf, outf1,outfn,outfn1,naftercon)
   else writeln(outf,'Not enough runs for meaningful results.');
 end; (* if not escape *)
  FinishGibbs(outf, outf1);
end; (* gibbsing *)

procedure MultiGibbs.randomstart;
(* calculate random values of theta, eta, x *)
var i,j,k: integer;
  w,pp: dvec;
  prpr: realvec;
begin
  for i:= cmax downto 1 do w[i] := 1.0;
(* Random theta : *)
  Dirichlet(c,w,pp);
  for i:=c downto 1 do theta[i] := pp[i];
(* Random eta : *)
  for i:=c downto 1 do begin
    for j:=i-1 downto 1 do begin
      Dirichlet(r,w,pp);
      for k:=r downto 1 do begin
        eta[k,j,i] := pp[k];
        eta[ppi[k],i,j] := pp[k];
       end; (* k *)
     end; (* j *)
    Dirichlet(r01,w,pp);
    for k:= r01 downto 1 do eta[k,i,i] := pp[k];
    for k:= r01+1 to r do eta[k,i,i] := pp[ppi[k]];
   end; (* i *)
(* Random x : *)
 for i:= n downto 1 do begin
   if identified and prior[i] then begin
     for k:=0 to cmax do prpr[k] := 0.0;
     for k:=1 to c do prpr[k] := thetai[i,k];
     x[i] := ranp(3,prpr);
    end else x[i] := 1 + randomd(3,c);
(* next 2 lines have probability 0 : *)
   if (x[i] < 1) then x[i] := 1;
   if (x[i] > c) then x[i] := c;
  end; (* i *)
end; (* randomstart *)

procedure MultiGibbs.optimalstart(var outf, outf1: textfile;
                                  outfn, outfn1: string);
(* This procedure defines the array x0 with starting values
   for the Gibbs sequences. *)
(* From version 1.6: calculate smarter starting values of theta, eta, x
  This is done by calculating sets of c vertices that have mutually
  the lowest numbers of matches (as indicator of similarity ),
  using these as nuclei for a new coloring,
  and defining the new initial coloring on the basis of
  the similarity of the other vertices to these nuclei. *)
var i,j,k,s,h,matchmin,nsep,nstarts,imax,himax: integer;
  color: array of array of integer;
  info, infoo: ConstnVec;
  infomin : real;
  imin, dismax, dismin: longword;
  distances: array[1..startmax div 2, 1..startmax div 2] of longword;
  index: array[1..startmax div 2] of integer;
  number: array[1..mulmax] of integer;
  logeta:     array[1..rmax, 1..cmax, 1..cmax] of real;
  match: array of array of integer;
  sep: array of integer;
  use: array of boolean;
  nuclei: array[1..cmax] of integer;
  found,selected: boolean;
function Rand(i,j: integer): longword;
(* Rand distance between partitions induced y colorings color[i,.]
   and color[j,.] *)
var r0: longword;
 h,k: integer;
begin
  r0 := 0;
  for h:=n downto 2 do for k:=h-1 downto 1 do
    if ((color[i,h] = color[i,k]) <> (color[j,h] = color[j,k]))
    then Inc(r0);
  Rand := r0;
end; (* Rand *)
begin (* OptimalStart *)
  writeln(outf,'Optimal start.');
  writeln(outf,'--------------'); writeln(outf);
  SetLength(color,startmax+1,n+1);
  InitGibbs(outf,outf1, outfn, outfn1);
  OptMessage1(startmax);
  infomin := 1e30;
  imin := 1;
  i := 0;
  if identified then Identifyvertices(outf);
(*  theta0 := sqr(n); *)
  theta0 := 1 + theta00*c;

 if randoptimize then begin
  (* First startmax random starting values *)
  repeat
    inc(i);
    RandomStart;
    calccounts;
    converged := false;
    w0 := 1.0;
    for j:=1 to 10 do onestep(false);
    informationy := 0.0;
    w0 := 100.0; (* for overdispersion *)
    onestep(true);
    info[i] := informationy/ndat;
    for j:=1 to n do color[i,j] := x[j];
    if (infomin > info[i]) then begin
      infomin := info[i];
      imin := i;
     end;
    OptMessage2(i,info[i],infomin);
   until ((i >= startmax) or Escape);
   imax := i;
 end else (* not randoptimize *) begin
(* this is a very long part, seek for next randoptimize comment *)
(* first determine number of well-separated vertices that have to be found *)
randomstart;
  k := 1;
  nsep := c;
  repeat
    inc(nsep);
    k := (k*nsep) div (nsep - c); (* = (nsep choose c ) *)
   until (k >= startmax);

  SetLength(match,n+1,n+1);
  SetLength(sep,nsep+1);
  SetLength(use,nsep+1);
  for i:=0 to nsep do sep[i] := 0;
  matchmin := n;
(* calculate numbers of matches as similarity coefficients : *)
  for i:=2 to n do begin
    for j := (i-1) downto 1 do begin
      s := 0;
      for k:=n downto 1 do if (y[i,k] = y[j,k]) then inc(s);
(* no check for data: missings do not contribute to similarity *)
      match[i,j] := s;
      match[j,i] := s;
      if (i <= 2) then begin
        matchmin := s;
        sep[1] := i;
        sep[2] := j;
       end else
        if (s < matchmin) then begin
        matchmin := s;
        sep[1] := i;
        sep[2] := j;
       end;
     end; (* for j *)
    match[i,i] := n;
   end; (* for i *)
  match[1,1] := n;
(* now sep[1] and sep[2] have the minimum similarity *)
  writeln(outf1,'matching numbers used for finding optimal start coloring: ');
  write(outf1,'    ');
  for i:=1 to n do begin
    if (n > 100) then write(outf1,(i mod 100):2,' ') else write(outf1,i:3,' ');
    if ((i mod 100) = 0) then
      begin writeln(outf1); write(outf1,(100*(i div 100)):4,' '); end;
   end;
  writeln(outf1);
  if (n >= 101) then writeln(outf1);
  for i:=1 to n do begin
    write(outf1,i:3,' ');
    for j:=1 to n do begin
      if (n > 100) then write(outf1,(match[i,j]div 10):2,' ') else write(outf1,match[i,j]:3,' ');
      if ((j mod 100) = 0) then
       begin writeln(outf1); write(outf1,'    '); end;
     end;
    if (n >= 101) then writeln(outf1);
   end; (* i *)
  writeln(outf1);
  writeln(outf1);

  for i:=3 to nsep do begin
    matchmin := n;
    for j:=n downto 1 do begin
      s := 0;
      for k := pred(i) downto 1 do if (match[j,sep[k]] > s)
        then s := match[j,sep[k]];
      if ((s < matchmin)) then begin
        matchmin := s;
        sep[i] := j;
       end;
     end; (* for j *)
(* now sep[i] has the smallest maximum similarity with sep[1]..sep[i-1] *)
   end; (* for i *)
  writeln(outf1,'Vertices with widest separation are :');
  for i:=1 to nsep do write(outf1,sep[i],' ');
  writeln(outf1); writeln(outf1);

  for i:=0 to c do use[i] := true;
  for i:=c+1 to nsep do use[i] := false;
  nstarts := 0;
  repeat
   inc(nstarts);
   if (nstarts >= 2) then begin
(* Find the next combination of c used nuclei out of the nsep available *)
     j := 0;
     found := false;
     repeat
       inc(j);
       if (use[j] and (not use[succ(j)])) then begin
         use[j] := false;
         use[succ(j)] := true;
         s := 0;
         for k:=1 to j do if use[k] then inc(s);
         for k:=1 to s do use[k] := true;
         for k:=s+1 to j-1 do use[k] := false;
         found := true;
        end;
      until (found or (j >= pred(nsep)));
     end; (* if *)
    i := 0;
    for j := 1 to nsep do if use[j] then begin
      inc(i);
      nuclei[i] := sep[j];
     end;
//  writeln(outf1,'Starting nuclei for coloring ',nstarts,' are :');
//  for i := 1 to c do write(outf1,nuclei[i],' ');
//  writeln(outf1);

(* Determine coloring x on the basis of best similarities with nuclei[i] : *)
    for i :=1 to n do begin
      s := 0;
//      write(outf1,i:3);
      for j:=1 to c do begin
//        write(outf1,' ',j,': ',match[i,nuclei[j]]:4,'; ');
        if (match[i,nuclei[j]] > s) then begin
          x[i] := j;
          s := match[i,nuclei[j]];
         end;
        end;
(* next 2 lines have probability 0 : *)
//      writeln(outf1,'color ',x[i]);
      if (x[i] < 1) then x[i] := 1;
      if (x[i] > c) then x[i] := c;
     end; (* for i *)
    calccounts;
    converged := false;
    w0 := 1.0;
    drawtheta;
    draweta(false);
    informationy := 0.0;
    w0 := 100.0; (* for overdispersion *)
//for i:=1 to n do write(outf1,x[i]);
//writeln(outf1);

     informationy := 0.0;
     for i:=c downto 1 do for j:=c downto 1 do
      for k :=r downto 1 do logeta[k,i,j] := log(eta[k,i,j]);
     for i:=n downto 2 do for j:=i-1 downto 1 do
      if data[i,j] then
      informationy := informationy + logeta[y[i,j],x[i],x[j]];
     informationy := -informationy/ndat;
//     for i:=1 to n do write(outf1,x[i]);
//writeln(outf1);
    for j:=1 to n do color[nstarts,j] := x[j];
    info[nstarts] := informationy;
    if (infomin > info[nstarts]) then begin
      infomin := info[nstarts];
      imin := nstarts;
//writeln(outf1,'infomin ',infomin:20:10);
//for i:=1 to n do write(outf1,x[i]);
//writeln(outf1);
     end;
//  writeln(outf1,'informationy ',informationy: 20:10, ' infomin: ',infomin:20:10);
    OptMessage2(nstarts,info[nstarts],infomin);
   until ((nstarts >= startmax) or Escape);
   Finalize(match);
   Finalize(sep);
   Finalize(use);
   imax := nstarts;
  end; (*  not randoptimize *)

  stop1:=false;
  informationy := 0.0;


(* Find coloring with smallest info *)
  for j:=1 to n do x0[1,j] := color[imin,j];
  for j:=1 to n do x[j] := x0[1,j];
  writeln(outf);
  writeln(outf,'Coloring with smallest information in ',imax,' trials is ');
  for j:=1 to n do write(outf,x[j]);
  writeln(outf,'      (information ',infomin:12:6,').');
  writeln(outf,'This is used as the initial coloring.');

//  for j:=1 to n do write(lf,x[j]);
//  writeln(lf);


  if multiple then begin
    writeln(outf);
    writeln(outf,'Multiple starting points.');
    writeln(outf,'-------------------------');
    writeln(outf);
(* Now obtain among the imax/2 colorings with the smallest info[i],
   multi colorings with large mutual Rand distances between all of them. *)
    infoo := info;
    quicksort(infoo,1,imax);
    himax := (2+imax) div 2;
    if (himax > (startmax div 2)) then himax := (startmax div 2);
    writeln(outf1,'smallest and largest information among ',
            imax,' random starting points are');
    writeln(outf1,infoo[1]:9:4,' and ',infoo[imax]:9:4);
    writeln(outf1,'starting points with information up to ',infoo[himax]:9:4,
                  ' are considered.');
    h := 1;
  (* Select himax colorings with smallest value of info *)
    index[1] := imin;
    for i:=1 to imax do
      if ((i <> imin) and
          (info[i] <= infoo[himax]) and (h < startmax div 2)) then begin
        inc(h);
        index[h] := i;
       end;
  (* calculate Rand distances *)
    for i:=1 to himax do begin
      for j:=1 to i-1 do begin
        distances[i,j] := Rand(index[i],index[j]);
        distances[j,i] := distances[i,j];
       end; (* j *)
      distances[i,i] := 0;
     end;

    number[1] := 1;
    if (multi <= 2) then
     writeln(outf,'Next coloring that is a well dispersed starting point:')
    else
     writeln(outf,'Next colorings that are well dispersed starting points:');
    writeln(outf);
    for i:=2 to multi do begin
      dismax := 0;
      imax := 1;
      for j:=1 to himax do begin
  (* calculate smallest distance between coloring j and
     colorings number[1] through number[i-1] *)
        selected := false;
        for h:=i-1 downto 1 do if (number[h] = j) then selected := true;
        if (not selected) then begin
          dismin := 32767;
          for h:=i-1 downto 1 do
           if (distances[j,number[h]] < dismin)
            then dismin := distances[j,number[h]];
(* select the largest of these smallest distances *)
          if (dismin > dismax) then begin
            dismax := dismin;
            imax := j;
           end; (* if *)
          end; (* if not selected *)
       end; (* j *)
      number[i] := imax;
      for j:=1 to n do x0[i,j] := color[index[number[i]],j];
      write(outf,i:4,'  ');
      for j:=1 to n do write(outf,x0[i,j]);
      writeln(outf,'      (information ',info[number[i]]:12:6,').');
      write(outf,'Rand distance to preceding coloring');
      if (i >= 3) then write(outf,'s');
      writeln(outf,' at least ',dismax,'.');
      writeln(outf);
     end; (* for i *)
   end; (* if multiple *)

  FinishGibbs(outf,outf1);
  theta0 := 1 + theta00*c;
  Finalize(color);
end; (* OptimalStart *)

procedure MultiGibbs.initgibbs(var outf, outf1: textfile; outfn, outfn1: string);
begin
 if (c > cmax) then begin
   writeln(outf,'The number of colors is ',c,'.');
   writeln(outf,'This is more than constant cmax = ', cmax,' in B_CONS.');
   writeln(outf,'Recompile BLOCKS with larger constant cmax.');
   stopsthis(outf, outf1);
   Halt;
  end; (* if *)

  EntrFrequency := 10;
  EntrShowFrequency := 5;
  ShowXEFrequency := 5;
  if (n <= 30) then width := 2 else width := 1;
  ntiny := 0;

  writeln(outf1);
  if identified then write(outf1,'I') else write(outf1,'Uni');
  writeln(outf1,'dentified model.');
  if identified then writeln(outf1,'-----------------')
    else writeln(outf1,'-------------------');
  writeln(outf1);
  (* if identified then write(outf,'I') else write(outf,'Uni');
  writeln(outf,'dentified model.');
  if identified then writeln(outf,'-----------------')
    else writeln(outf,'-------------------'); *)
  (* writeln(outf,'Maximum possible entropy for color distribution is ',
                  log(c):7:4); *)

end; (* initgibbs *)

procedure MultiGibbs.calccounts;
(* calculates values of m,e,d *)
var i,j,k: integer;
begin
(* Initialise with 0 values *)
  for i:=r downto 1 do
   for j:=c downto 1 do begin
     for k:=c downto 1 do e[i,j,k] := 0;
     for k:=n downto 1 do d[i,k,j] := 0;
    end; (* i,j *)
  for i:=c downto 1 do
    begin m[i] := 0; mm[i] := 0; end;

(* Calculate counts *)
  for i:=n downto 1 do begin
      inc(mm[x[i]]);
      if (not prior[i]) then inc(m[x[i]]);
    for j:=n downto 1 do if data[i,j] then begin
      inc(e[y[i,j],x[i],x[j]]);
      inc(d[y[i,j],i,x[j]]);
     end;
   end; (* i *)
  for i:=c downto 1 do
   for k:=r0 downto 1 do e[k,i,i] := e[k,i,i] div 2;
end; (* calccounts *)

procedure MultiGibbs.calcstats;
var i,j,k: integer;
begin
 inc(nstat);
 if identified then begin
   for i:=c downto 1 do begin
     thetax[i] := thetax[i] + theta[i];
     theta2[i] := theta2[i] + sqr(theta[i]);
    end;
   for i:=1 to n do begin
     predm[i] := predm[i] + mm[x[i]];
     classprob[i,x[i]] := classprob[i,x[i]] + 1.0;
     for j:=i-1 downto 1 do
       if (x[i] = x[j]) then xpair[i,j] := xpair[i,j] + 1.0;
    end; (* i *)
   for k:=r01 downto 1 do for i:=c downto 1 do for j:=c downto 1 do begin
     etac[k,i,j] := etac[k,i,j] + eta[k,i,j];
     if calcsd then etac2[k,i,j] := etac2[k,i,j] + sqr(eta[k,i,j]);
    end; (* k *)
  end (* if *) else (* not identified *)
   for i:=1 to n do begin
     thetax[i] := thetax[i] + theta[x[i]];
     predm[i] := predm[i] + mm[x[i]];
     for j:=i-1 downto 1 do
       if (x[i] = x[j]) then xpair[i,j] := xpair[i,j] + 1.0;
     for j:=n downto 1 do if (i <> j) then
       for k:=r01 downto 1 do begin
         etax[k,i,j] := etax[k,i,j] + eta[k,x[i],x[j]];
         if calcsd then
           etax2[k,i,j] := etax2[k,i,j] + sqr(eta[k,x[i],x[j]]);
        end;
    end; (* i *)
end; (* calcstats *)

procedure MultiGibbs.drawx(CalcEntropy: Boolean);
(* draw new values of x and update e,d,m,mm *)
const tiny = 1e-300;
{      nmax = ???; }
var pp: realvec;
 aa,k,l,i,j,xi0,xi1: integer;
 s,entx: real;
(* for checking only : *)
{m0:          array[1..nmax] of integer;
f0:          array[1..rmax,1..nmax,1..cmax] of integer;
e0:          array[1..rmax,1..cmax,1..cmax] of longword;
errors: boolean; }
begin
(* Initialise *)
    for j:= 0 to cmax do pp[j] := 0.0;
    entx := 0.0;
  for i:=1 to n do begin
    xi0 := x[i];

(* First compute the probability vector pp *)
    s := 0.0;
    for k:=c downto 1 do begin
      if (identified and prior[i])
        then pp[k] := thetai[i,k] else pp[k] := theta[k];
      for aa:= r downto 1 do begin
        for l := c downto 1 do pp[k] := pp[k] * power(eta[aa,k,l],d[aa,i,l]);
(*  write(lf,'k = ',k ,',aa = ',aa,' ');
  for l:=c downto 1 do writeln(lf,theta[k],' '); writeln(lf);
  for l:=c downto 1 do writeln(lf,d[aa,i,l],' '); writeln(lf);
  for l:=c downto 1 do writeln(lf,log(eta[aa,k,l])*d[aa,i,l],' '); writeln(lf);
  for l:=c downto 1 do writeln(lf,power(eta[aa,k,l],d[aa,i,l]),' '); writeln(lf);
  *)
       end; (* aa *)
      s := s+pp[k];
     end; (* k *)
//    for k:=1 to c do write(lf,pp[k]:7:4,' '); writeln(lf);
    if (s > tiny) then begin
      s := 1/s;
      for k:=c downto 1 do pp[k] := pp[k]*s;
      end
     else begin
       inc(ntiny);
       for k:=c downto 1 do pp[k] := 1.0/c;
    (* Probability that s <= tiny is negligible, unless there are
       numerical problems *)
      end;

(* Now draw the new x[i] *)
    x[i] := ranp(3,pp);
    if (x[i] < 1) then x[i] := 1;   (* this has probability 0 *)
    if (x[i] > c) then x[i] := c;   (* this too *)

{    write(lf,i:3,'.');
    for k:=1 to c do write(lf,pp[k]:7:4,' ');
    writeln(lf,':  ',x[i]);
 }
   if CalcEntropy then for k:=c downto 1 do entx := entx + pp[k]*log(pp[k]);

(* Update m,mm,e,d if x[i] has changed *)
    if (x[i] <> xi0) then begin
      xi1 := x[i];
      Dec(mm[xi0]);
      Inc(mm[xi1]);
      if (not prior[i]) then begin
        Dec(m[xi0]);
        Inc(m[xi1]);
       end; (* if *)
      for j:=n downto 1 do if data[i,j] then begin
        Dec(d[y[j,i],j,xi0]);
        Inc(d[y[j,i],j,xi1]);
        Dec(e[y[i,j],xi0,x[j]]);
        if ((y[j,i] > r0) or (x[j] <> xi0)) then Dec(e[y[j,i],x[j],xi0]);
        Inc(e[y[i,j],xi1,x[j]]);
        if ((y[j,i] > r0) or (x[j] <> xi1)) then Inc(e[y[j,i],x[j],xi1]);
       end; (* for j  *)
     end; (* if x[i] <> xi0 *)
   end; (* i *)

  if CalcEntropy then begin
    inc(NEntropy);
    Entropyx := Entropyx - entx;
   end; (* if *)

{
(* To check the updates : *)
  for aa := 1 to r do for k:= 1 to c do begin
    for i:= 1 to n do f0[aa,i,k] := d[aa,i,k];
    for i:= 1 to c do e0[aa,i,k] := e[aa,i,k];
   end; (* aa, k *)
  for i:=1 to c do m0[i] := m[i];
  errors := false;
  calccounts;
  for aa := 1 to r do for k:= 1 to c do begin
    for i:= 1 to n do if f0[aa,i,k] <> d[aa,i,k] then begin
      errors := true;
      writeln(lf, 'error d[',aa,',',i,',',k,'] : ',f0[aa,i,k],' <> ',d[aa,i,k]);
     end;
    for i:= 1 to c do if e0[aa,i,k] <> e[aa,i,k] then begin
      errors := true;
      writeln(lf, 'error e[',aa,',',i,',',k,'] : ',e0[aa,i,k],' <> ',e[aa,i,k]);
     end;
   end; (* aa, k *)
  for i:=1 to c do if m0[i] <> m[i] then begin
    errors := true;
    writeln(lf, 'error m[',i,'] : ',m0[i],' <> ',m[i]);
   end; (* i *)
  if errors then begin
    writeln(lf, 'errors');
   end;         }

end; (* drawx *)

procedure MultiGibbs.drawtheta;
var i: integer;
  w,pp: dvec;
begin
  for i:=c downto 1 do w[i] := m[i] + theta0;
  Dirichlet(c,w,pp);
  for i:=c downto 1 do theta[i] := pp[i];
end; (* drawtheta *)

procedure MultiGibbs.draweta(CalcEntropy: Boolean);
var i,j,aa: integer;
 w,pp: dvec;
 logeta:     array[1..rmax, 1..cmax, 1..cmax] of real;
 infoy: real;
begin
  for i:=c downto 1 do begin
    for j:=i-1 downto 1 do begin
      for aa:=r downto 1 do w[aa] := w0*(e[aa,i,j] + eta0[aa,i,j]);
      Dirichlet(r,w,pp);
      for aa:=r downto 1 do begin
        eta[aa,i,j] := pp[aa];
        eta[ppi[aa],j,i] := pp[aa];
       end; (* k *)
     end; (* j *)
    for aa:=r01 downto 1 do w[aa] := w0*(e[aa,i,i] + eta0[aa,i,i]);
    Dirichlet(r01,w,pp);
    for aa:= r0 downto 1 do eta[aa,i,i] := pp[aa];
    for aa:= r1 downto 1 do begin
      eta[aa+r0,i,i] := 0.5*pp[aa+r0];
      eta[aa+r0+r1,i,i] := 0.5*pp[aa+r0];
     end; (* aa *)
   end; (* i *)

(* Now for information calculation *)
   if CalcEntropy then begin
     infoy := 0.0;
     for i:=c downto 1 do for j:=c downto 1 do
      for aa :=r downto 1 do logeta[aa,i,j] := log(eta[aa,i,j]);
     for i:=n downto 2 do for j:=i-1 downto 1 do
      if data[i,j] then infoy := infoy + logeta[y[i,j],x[i],x[j]];
     informationy := informationy - infoy;
    end;

end; (* draweta *)


procedure MultiGibbs.GibbsResults(var outf, outf1: textfile;
                      outfn, outfn1: string; naftercon: integer);
var i,j,k,h,mini: integer;
 col,rank,ar: array of integer;
 width0: integer;
 maxprob,minprob,sr,sx,sx1: real;
 threshold: array[0..cmax] of real;
// chosen,
 last: array of boolean;
(*  last: indicator of last vertex of its color in StricBlocks *)
function fetax(k,i,j:integer):real;
begin
  if (k <= r01) then fetax := etax[k,i,j]
   else fetax := etax[ppi[k],j,i];
end; (* fetax *)
function fetax2(k,i,j:integer):real;
begin
  if (k <= r01) then fetax2 := etax2[k,i,j]
   else fetax2 := etax2[ppi[k],j,i];
end; (* fetax2 *)
function fetac(k,i,j:integer):real;
begin
  if (k <= r01) then fetac := etac[k,i,j]
   else fetac := etac[ppi[k],j,i];
end; (* fetac *)
function fetac2(k,i,j:integer):real;
begin
  if (k <= r01) then fetac2 := etac2[k,i,j]
   else fetac2 := etac2[ppi[k],j,i];
end; (* fetac2 *)
procedure head(k: byte; ss: string);
var i,h: integer;
begin
(* main/rough results on outf *)
  if (ss > '') then begin
    writeln(outf);
    writeln(outf,ss);
   end;
  if (n >= 1000) then begin
    write(outf,'      ');
    for i:=1 to n do begin
      if ((ar[i] div 1000) = 0) then write(outf,' ')
       else write(outf,(ar[i] div 1000):1);
      if last[i] then write(outf,'.');
     end;
    writeln(outf);
   end;
  if (n >= 100) then begin
    write(outf,'      ');
    for i:=1 to n do begin
      if ((ar[i] div 100) = 0) then write(outf,' ')
       else write(outf,((ar[i] div 100) mod 10):1);
      if last[i] then write(outf,'.');
     end;
    writeln(outf);
   end;
  write(outf,'      ');
  for i:=1 to n do begin
    if ((ar[i] div 10) = 0) then for h:=1 to width do write(outf,' ')
     else write(outf,((ar[i] div 10) mod 10):width);
    if last[i] then
      if (width <= 1) then write(outf,'.') else write(outf,' . ');
   end;
  writeln(outf);
  write(outf,'      ');
  for i:=1 to n do begin
    write(outf,(ar[i] mod 10):width);
    if last[i] then
      if (width <= 1) then write(outf,'.') else write(outf,' . ');
   end;
  writeln(outf); writeln(outf);
(* more detailed on outf1 *)
 if ((k >= 2) and (n <= 100)) then begin
  if (ss > '') then begin
    writeln(outf1);
    writeln(outf1,ss);
   end;
  if (n >= 1000) then begin
    write(outf1,'      ');
    for i:=1 to n do begin
      if ((ar[i] div 1000) = 0) then write(outf1,'   ')
       else write(outf1,(ar[i] div 1000):3);
      if last[i] then write(outf1,'...');
     end;
    writeln(outf1);
   end;
  if (n >= 100) then begin
    write(outf1,'      ');
    for i:=1 to n do begin
      if ((ar[i] div 100) = 0) then write(outf1,'   ')
       else write(outf1,((ar[i] div 100) mod 10):3);
      if last[i] then write(outf1,'...');
     end;
    writeln(outf1);
   end;
  write(outf1,'      ');
  for i:=1 to n do begin
    if ((ar[i] div 10) = 0) then for h:=1 to width do write(outf1,'   ')
     else write(outf1,((ar[i] div 10) mod 10):3);
    if last[i] then write(outf1,'...');
   end;
  writeln(outf1);
  write(outf1,'      ');
  for i:=1 to n do begin
    write(outf1,(ar[i] mod 10):3);
    if last[i] then write(outf1,'...');
   end;
  writeln(outf1); writeln(outf1);
 end; (* if k >= 2 *)
end; (* head *)
procedure writed(xx: real;numf: integer);
var ix,ix1: integer;
begin
  ix := round(xx*10);
  ix1 := round(xx*100);
  if (ix >= 10) then ix := 9;
  if (numf >= 2) then write(outf,ix:width);
  if (n <= 100) then write(outf1,ix1:3);
end; (* writed *)

(* Revised version T.S. November 27, 2000;
  differences between Blocks versions 1.4 and 1.41:
  replace StrictBlock in Mulblo.Pas by the code below; *)

procedure Strictblock;
(* In the output phase, this procedure calculates and outputs
   a posterior block model on the basis of xpair, according to a strict
   criterion which can involve deleting vertices of which the
   color is not clearly defined *)
var i,j,h,k,imost,outcount,outcount1,firsti,il,cols,colsm,imm: integer;
  v: array of integer; (* first, number of violations per vertex;
                                   later, color of vertex *)
  throwout: array of boolean;
      (* throwout: indicators of deletion of vertices *)
  ntot: array[1..cmax,1..cmax] of integer;
  max_dif, min_same, max_del, min_del: array of real;
    (* max of xpair for vertices with different color,
       in which this vertex participates;
       min of xpair for vertices with same color,
       in which this vertex participates;
       the same max, overall, if this vertex is deleted;
       the same min, overall, if this vertex is deleted;*)
  tr1, tr2, maxmax, minmin, mmdel: real; (* thresholds, extremes *)
  rednumber,vmax, col0: integer;
  threeway: array[1..rmax,1..cmax,1..cmax] of real;
  pf: textfile; pn: shortstring;
begin (* MultiGibbs.StrictBlock *)
  SayResults2;
  Heading(outf,4,'Finding strictly separated classes.');
  SetLength(v, n+2);
  SetLength(throwout, n+2);
  SetLength(max_dif, n+2);
  SetLength(min_same, n+2);
  SetLength(max_del, n+2);
  SetLength(min_del, n+2);

  tr1 := 0.5;
  for i:=1 to n do begin
    throwout[i] := false; end;
  outcount := 0;
(* throw vertices out to obtain a partition of the remaining vertices
   in the following way:
   vertices i,j are supposed to have the same color if xpair[i,j] > tr1
   and a different color if xpair[i,j] < tr1;
   (i,j) has a triadic violation with k if xpair(i,j) < tr1
   so that i and j have not the same color, and
   (color(i) = color(k) and color(j) = color(k)).
   v[i] is the number of triadic violations in which vertex i is involved *)
  writeln(outf,'Vertices now can be thrown out because they are not compatible');
  writeln(outf,'with a partition of vertices into colors.');
  repeat
   ShowWorking('x');
   for i:=1 to n do v[i] := 0;
   for i:=n downto 1 do for j:=n downto 1 do
     if ((i <> j) and (not throwout[i]) and (not throwout[j]) and
         (xpair[i,j] < tr1)) then begin
       for k:=n downto 1 do if (not throwout[k]) then
         if ((xpair[i,k] > tr1) and (xpair[j,k] > tr1)) then
           begin inc(v[i]); inc(v[j]); inc(v[k]); end;
      end;
   vmax := 0; imost := 0;
   for i:=1 to n do
    if (v[i] > vmax) then begin vmax := v[i]; imost := i; end;
   if (vmax > 0) then
     begin
       throwout[imost] := true; inc(outcount);
      end;
   until ((vmax = 0) or Escape);
  writeln(outf);
 (* Now the vertices that are not out are partitioned with threshold tr1 *)


  outcount1 := outcount;
  tr1 := 0.5; tr2 := 0.5;
  rednumber := 0;
  repeat
(* This repeat runs on until
     ((minmin - maxmax >= 0.6) or (outcount > n div 5)) *)
   ShowWorking('.');
   if (outcount > 0) then inc(rednumber);
   for i:=1 to n do last[i] := false;
 (* subsequently throw out poorly fitting vertices *)
 (* Count number of colors cols:
    give colors v[ar[i]] subsequently to the vertices from ar[1] to ar[n] *)
    cols := 0; (* current color *)
    for i:=1 to n do if (throwout[i]) then v[i] := -1 else v[i] := 0;
    firsti := 1;
    repeat
      inc(cols);
      v[ar[firsti]] := cols;
      imost := n+1;
      for i:= n downto (firsti + 1) do
        if (not throwout[ar[i]]) then begin
          if (xpair[ar[firsti],ar[i]] > tr2) then v[ar[i]] := cols
           else imost := i;
         end;
      firsti := imost;
     until (firsti > n);
 (* Calculate max_dif, min_same, max_del, min_del, as defined above: *)
   for i:=1 to n do
     begin max_dif[i] := 0.0; min_same[i] := 1.0;
           max_del[i] := 0.0; min_del[i] := 1.0; end;
   for i:=n-1 downto 1 do for j:=i+1 to n do
     if ((not throwout[i]) and (not throwout[j])) then begin
       if (xpair[i,j] < tr1) then begin
         if (xpair[i,j] > max_dif[i]) then max_dif[i] := xpair[i,j];
         if (xpair[i,j] > max_dif[j]) then max_dif[j] := xpair[i,j];
          for k:=n downto 1 do if ((k <> i) and (k <> j) and
            (xpair[i,j] > max_del[k])) then max_del[k] := xpair[i,j];
        end;
     if (xpair[i,j] > tr2) then begin
       if (xpair[i,j] < min_same[i]) then min_same[i] := xpair[i,j];
       if (xpair[i,j] < min_same[j]) then min_same[j] := xpair[i,j];
        for k:=n downto 1 do if ((k <> i) and (k <> j) and
          (xpair[i,j] < min_del[k])) then min_del[k] := xpair[i,j];
         end;
      end; (* for i, j, if not throwout *)
   maxmax := 0.0; minmin := 1.0; mmdel := 0.0; imm := 0;
   for i:=1 to n do if (not throwout[i]) then begin
     if (max_dif[i] > maxmax) then maxmax := max_dif[i];
     if (min_same[i] < minmin) then minmin := min_same[i];
     if ((min_del[i] - max_del[i]) > mmdel) then begin
       mmdel := min_del[i] - max_del[i];
       imm := i;
      end;
     end; (* if not throwout *)

   if ((imm >= 1) and (minmin - maxmax < 0.6))  then begin
     throwout[imm] := true; inc(outcount);
    end;

(* determine the last vertex of the various colors *)
   col0 := 0; i := 0; il := 0;
   repeat
     inc(i);
     if (not throwout[ar[i]]) then begin
       if (v[ar[i]] <> col0) then begin
         if (col0 >= 1) then last[il] := true;
         col0 := v[ar[i]];
        end;
      il := i;
      end;
    until (i >= n);

    (*  min_del[i] - max_del[i], which is positive for all i,
        indicates how much the fit is improved by throwing away vertex i *)
    if ((rednumber = 1) or (outcount = outcount1) or (minmin - maxmax >= 0.6)
            or (outcount > (n div 5)) or Escape)
     (* this condition for output is supposed to
        hold the first and the last time when passing
        this repeat .. until loop *)
     then begin
      writeln(outf);
      if ((rednumber = 1)  and (outcount = 0)) then begin
        Heading(outf,5,'Complete vertex set');
       end else begin
          Heading(outf,5,'Reduced vertex set');
          writeln(outf,'The following vertices were thrown out because of poor fit in the block structure:');
          for i:=1 to n do if throwout[i] then write(outf,i,' ');
          writeln(outf,'.');
       end;
      writeln(outf);
      writeln(outf,
       'Overall maximum pairwise value for vertices in different classes is ',
        maxmax:5:3);
      writeln(outf,
       'Overall minimum pairwise value for vertices in the same class is    ',
        minmin:5:3);
      writeln(outf);
      writeln(outf,'Classes now are labeled 1 to ',cols,'.');
      writeln(outf,'Vertices, as far as not thrown out, with their class number,');
      writeln(outf,'their maximum pairwise value for vertices in a different class,');
      writeln(outf,'and their minimum pairwise value for vertices in the same class;');
      writeln(outf,'and, if this vertex were to be deleted:');
      writeln(outf,'the overall maximum pairwise value for vertices in different classes,');
      writeln(outf,'and the overall minimum pairwise value for vertices in the same class:');
      writeln(outf);
      for i:=1 to n do begin
        write(outf,i:4);
        if throwout[i] then writeln(outf,' out ')
         else writeln(outf,v[i]:3,'  ',max_dif[i]:5:3, min_same[i]:7:3,
                                       max_del[i]:7:3, min_del[i]:7:3);
       end;
      writeln(outf);
      writeln(outf,'In new ranking:');
      for i:=1 to n do begin
        write(outf,ar[i]:4);
        if throwout[ar[i]] then writeln(outf,' out ')
         else writeln(outf,v[ar[i]]:3,'  ',max_dif[ar[i]]:5:3, min_same[ar[i]]:7:3,
                                   max_del[ar[i]]:7:3, min_del[ar[i]]:7:3);
       end;
      writeln(outf);

//      if (n <= 80) then begin

      writeln(outf,'Recoded adjacency matrix with block structure :');
      writeln(outf);
      Head(1,'');
      for i:=1 to n do begin
        write(outf,ar[i]:4,'  ');
        if (not throwout[ar[i]]) then  for j:=1 to n do begin
          if ((i <> j) and (not throwout[ar[j]]))
            then write(outf,y[ar[i],ar[j]]:width) else
              for h:=1 to width do write(outf,' ');
          if last[j] then
            if (width <= 1) then write(outf,'.') else write(outf,' . ');
         end; (* j *)
        writeln(outf);
        if (last[i]) then begin
          write(outf,'  ');
          for j:=2 to cols do for h:=(2*width -1) downto 1 do write(outf,'.');
          write(outf,'....');
          for j:= 1 to n do for h:=1 to width do write(outf,'.');
          writeln(outf);
         end;
       end; (* i *)
      writeln(outf);

//      end; (* 80 *)

      Heading(outf,6,'Posterior probabilities and frequencies for post-hoc coloring');
      writeln(outf);
      if (cols > cmax) then begin
        writeln(outf,'The number of vertices in this partition is ',cols,',');
        writeln(outf,'larger than constant cmax = ',cmax,'.');
        writeln(outf,'For this reason, the following vertices are not');
        writeln(outf,'taken into consideration in these posterior results:');
        writeln(outf);
        writeln(outf,'vertex   color ');
        for i:=1 to n do if (v[i] > cmax) then writeln(outf,i:4,v[i]:8);
        writeln(outf);
        colsm := cmax;
        writeln(outf,'For classes 1 - ',colsm,', two threeway tables :');
       end (* if cols > cmax *) else begin
        writeln(outf,'For these classes, two threeway tables :');
        colsm := cols;
       end;
      writeln(outf);
    (* sum posterior probabilities in blocks: *)
      writeln(outf,'Posterior probabilities');
      writeln(outf,'-----------------------');
      for i:=1 to cmax do for j:=1 to cmax do begin
        ntot[i,j] := 0;
        for k:=1 to r do threeway[k,i,j] := 0.0;
       end;
      for i:=n downto 1 do if (not throwout[i]) then
        for j:=n downto 1 do if
          ((not throwout[j]) and (i <> j) and (v[i] <= cmax) and (v[j] <= cmax))
          then begin
        inc(ntot[v[i],v[j]]);
        for k:=1 to r do threeway[k,v[i],v[j]] :=
                         threeway[k,v[i],v[j]] + fetax(k,i,j);
       end; (* for i,j if i <> j *)

      for k:=1 to r do begin
        writeln(outf,
         'Average posterior probabilities of tie value '+st(k)+' = '+symbol(k)+' in these blocks.');
        write(outf,' '); for j:=1 to colsm do write(outf,j:8); writeln(outf);
        for i:=1 to colsm do begin
          write(outf,i:3);
          for j:=1 to colsm do begin
(* here was a range check error in version 1.52;
   solved by use of variable colsm *)
            if (ntot[i,j] > 0) then write(outf,threeway[k,i,j]/ntot[i,j]:7:3)
             else write(outf,'    .0 ');
           end;
          writeln(outf);
         end; (* i *)
        writeln(outf);
       end; (* for k *)

    (* count frequencies in blocks: *)
      writeln(outf);
      writeln(outf,'Observed frequencies');
      writeln(outf,'--------------------');
      for i:=1 to colsm do for j:=1 to colsm do begin
        ntot[i,j] := 0;
        for k:=1 to r do threeway[k,i,j] := 0.0;
       end;
      for i:=n downto 1 do if (not throwout[i]) then
        for j:=n downto 1 do if
          ((not throwout[j]) and (data[i,j]) and (v[i] <= cmax) and (v[j] <= cmax))
          then begin
        inc(ntot[v[i],v[j]]);
        threeway[y[i,j],v[i],v[j]] := threeway[y[i,j],v[i],v[j]] + 1.0;
       end; (* for i,j if i <> j *)

      for k:=1 to r do begin
        writeln(outf,
        'Observed relative frequencies of tie value '+st(k)+' = '+symbol(k)+' in these blocks.');
        write(outf,' '); for j:=1 to colsm do write(outf,j:7); writeln(outf);
        for i:=1 to colsm do begin
          write(outf,i:3);
          for j:=1 to colsm do
            if (ntot[i,j] > 0) then write(outf,threeway[k,i,j]/ntot[i,j]:7:3)
             else write(outf,'    .0 ');
          writeln(outf);
         end; (* i *)
        writeln(outf);
       end; (* for k *)

      if ((outcount > 0) and (outcount <= 50)) then begin
        writeln(outf,'Relations of poorly fitting vertices with these classes : ');
        writeln(outf,'--------------------------------------------------------- ');
        for i:=1 to n do if throwout[i] then begin
          for j:=1 to colsm do begin
            ntot[1,j] := 0;
            for k:=1 to r do threeway[k,1,j] := 0.0;
           end;
          for j:=n downto 1 do if
            ((not throwout[j]) and (i <> j) and (v[j] <= cmax))
            then begin
            inc(ntot[1,v[j]]);
            for k:=1 to r do threeway[k,1,v[j]] :=
                         threeway[k,1,v[j]] + fetax(k,i,j);
           end; (* for j if i <> j *)

          writeln(outf,
           'Average posterior probabilities of relations between vertex ',
            i, ' and these classes.');
          writeln(outf);
          write(outf,'        ');
          for j:=1 to colsm do write(outf,j:7); writeln(outf);
          for k:=1 to r do begin
            write(outf,symbol(k),' ');
            for j:=1 to colsm do
              if (ntot[1,j] > 0) then write(outf,threeway[k,1,j]/ntot[1,j]:7:3)
               else write(outf,'    .0 ');
            writeln(outf);
           end; (* r *)
          writeln(outf);

          for j:=1 to colsm do begin
            ntot[1,j] := 0;
            for k:=1 to r do threeway[k,1,j] := 0.0;
           end;
          for j:=n downto 1 do if
            ((not throwout[j]) and (data[i,j]) and (v[j] <= cmax))
            then begin
            inc(ntot[1,v[j]]);
            threeway[y[i,j],1,v[j]] := threeway[y[i,j],1,v[j]] + 1.0;
           end; (* for j if i <> j *)

          writeln(outf,
            'Observed relative frequencies of relations between vertex ',
             i, ' and these classes.');
          writeln(outf);
          write(outf,'        ');
          for j:=1 to colsm do write(outf,j:7); writeln(outf);
          for k:=1 to r do begin
            write(outf,symbol(k),' ');
            for j:=1 to colsm do
              if (ntot[1,j] > 0) then write(outf,threeway[k,1,j]/ntot[1,j]:7:3)
               else write(outf,'    .0 ');
            writeln(outf);
           end; (* r *)
          writeln(outf);
         end; (* for i if throwout[i] *)
       end; (* if outcount > 0 and <= 50 *)
      end; (* if conditions for output are satisfied *)

   until ((minmin - maxmax >= 0.6) or (outcount > (n div 5)) or Escape);

(* New in version 1.7: Pajek output *)
  pn := proname+'_r.paj';
  assignfile(pf,pn);
  if (clusternumber <= 0) then
    begin
    rewrite(pf);
    writeln(pf,'*Network ',proname);
    writeln(pf,'*Vertices ',n);
    for i:=1 to n do writeln(pf,i:4,' ',char(34),i,char(34));
    writeln(pf,'*Arcs');
    for i:=1 to n do for j:=1 to n do
      if (i <> j) and (oldcode1[y[i,j]] <> 0)
       then writeln(pf,i:4,' ',j:4,' ',oldcode1[y[i,j]]:3);
    end else
      begin
      append(pf);
      end;
  writeln(pf);
  inc(clusternumber);
  writeln(pf,'*Partition ',clusternumber,' ',c,' colors, H_x =', sx:7:4 );
  writeln(pf,'*Vertices ',n);
  for i:=1 to n do
    if ((v[i] >= 1) and (v[i] <= cmax)) then
     writeln(pf,' ',v[i]:3)
     else writeln(pf,'   0');
  closefile(pf);
  writeln(outf,'This post-hoc coloring is partition ',clusternumber,' in Pajek file ',pn,'.');

  Finalize(v);
  Finalize(throwout);
  Finalize(max_dif);
  Finalize(min_same);
  Finalize(max_del);
  Finalize(min_del);
end; (* StrictBlock *)

begin (* MultiGibbs.GibbsResults *)
  SayResults1;
  if (ntiny > naftercon div 10) then begin
    writeln(outf);
    writeln(outf,'*************************************************************');
    writeln(outf,'There are problems with numerical accuracy.');
    writeln(outf,'In ',ntiny,' cases, truncation of tiny numbers was necessary.');
    writeln(outf,'This is related to the use of very large numbers,');
    writeln(outf,'in view of the big data set.');
    writeln(outf,'The results are unreliable.');
    writeln(outf,'If you wish to analyse this data set using BLOCKS,');
    writeln(outf,'it must be programmed in greater numerical precision.');
    writeln(outf,'*************************************************************');
    writeln(outf);
   end;
  SetLength(col,n+3);
  SetLength(rank,n+3);
  SetLength(ar,n+3);
//  SetLength(chosen,n+3);
  SetLength(last,n+3);
  for i:=1 to n do last[i] := false;
  if (nentropym > 0) then begin
    entropyxm := entropyxm / nentropym;
    informationym := informationym / nentropym;
(*    write(outf,'Entropy x : ',entropyxm:8:4,'; '); *)
    writeln(outf,'Information y : ',informationym:8:4);
    writeln(outf1,'Information y : ',informationym:8:4);
    writeln(lf,'Information y : ',informationym:8:4);
   end;
(* Normalize statistics from sums to averages *)
 sx := 0.0; sx1 := 0.0;
 if identified then begin
   for i:=c downto 1 do begin
     thetax[i] := thetax[i]/nstat;
     theta2[i] := root((theta2[i]/nstat) - sqr(thetax[i]));
    end;
   for i:=1 to n do begin
     for j:=c downto 1 do classprob[i,j] := classprob[i,j]/nstat;
     for j:=i-1 downto 1 do begin
       xpair[i,j] := xpair[i,j]/nstat;
       xpair[j,i] := xpair[i,j];
       sx1 := sx1 + xpair[i,j];
       sx := sx + 4*xpair[i,j]*(1.0-xpair[i,j]);
      end; (* j *)
    end; (* for i *)
   for k:=r01 downto 1 do for i:=c downto 1 do for j := c downto 1 do begin
     etac[k,i,j] := etac[k,i,j]/nstat;
     if calcsd then
       etac2[k,i,j] := root((etac2[k,i,j]/nstat) - sqr(etac[k,i,j]));
    end; (* for k,i,j *)
  end (* if *) else (* not identified *)
   for i:=1 to n do begin
     thetax[i] := thetax[i]/nstat;
     for j:=i-1 downto 1 do begin
       xpair[i,j] := xpair[i,j]/nstat;
       xpair[j,i] := xpair[i,j];
       sx1 := sx1 + xpair[i,j];
       sx := sx + 4*xpair[i,j]*(1.0-xpair[i,j]);
      end; (* j *)
    for j:=n downto 1 do if (i <> j) then begin
       for k:=r01 downto 1 do begin
         etax[k,i,j] := etax[k,i,j]/nstat;
         if calcsd then
           etax2[k,i,j] := root((etax2[k,i,j]/nstat) - sqr(etax[k,i,j]));
        end; (* k *)
      end; (* j *)
    end; (* for i, not identified *)
  for i:=1 to n do predm[i] := predm[i]/nstat;
  if identified then begin
(* Get new order in accordance with latent classes *)
    for i:=1 to n do begin
      col[i] := 1; maxprob := classprob[i,1];
      for j:=2 to c do if (classprob[i,j] > maxprob) then begin
        col[i] := j;
        maxprob := classprob[i,j];
       end; (* if *)
     end; (* i *)
    for i:=1 to n do rank[i] := 0;
    k := 0;
    for j:=1 to c do
      for i:=1 to n do if (col[i] = j) then begin
        inc(k);
        rank[i] := k;
       end; (* if *)
  end else (* not identified *) begin
(* Get new order that indicates approximate latent classes *)
(* k is the last rank assigned *)
    for i:=1 to n do rank[i] := 0;
    k:= 0;
    mini := 1;
    repeat
(* mini is the smallest vertex number
   for which a rank number has not yet been assigned *)
      inc(k);
      rank[mini] := k;
(* Now mini is the vertex number of which the class is to be found *)
       for j:=mini+1 to n do
       if ((rank[j] = 0) and (xpair[mini,j] > 0.5)) then begin
         inc(k);
         rank[j] := k;
        end; (* if *)
      while ((rank[mini] > 0) and (mini <= n)) do inc(mini);
     until (mini > n);
(* Determine base vertices for partition in c colors *)
{     for i:=1 to n do chosen[i] := false;
     minprob := 2.0; colver[1] := 1; colver[2] := 2;
     for i:=1 to n-1 do for j:=i+1 to n do
       if (xpair[i,j] < minprob) then begin
         minprob := xpair[i,j];
         colver[1] := i; colver[2] := j;
        end;
     chosen[colver[1]] := true; chosen[colver[2]] := true;
     threshold[1] := 0.0;
     threshold[2] := minprob;
     for k:=3 to c do begin
       minprob := 2.0; colver[k] := 3;
       for i:=1 to n do if (not chosen[i]) then begin
         maxprob := 0.0;
         for j:= k-1 downto 1 do
           if ((i <> colver[j]) and (xpair[i,colver[j]] > maxprob))
             then maxprob := xpair[i,colver[j]];
         if (maxprob < minprob) then begin
           minprob := maxprob;
           colver[k] := i;
          end; (* if < *)
        end; (* if not chosen i *)
       chosen[colver[k]] := true;
       threshold[k] := minprob;
      end; (* for k *)
     lastthreshold := threshold[c];
  }

   minprob := 1.0;
   for i:=n downto 1 do for j:=n downto 1 do
     if ((i <> j) and (xpair[i,j] < minprob)) then begin
       colver[1] := i;
       colver[2] := j;
       minprob := xpair[i,j];
      end;
   threshold[2] := minprob;
   minprob := 1.0;
   for i:=3 to c do begin
    for j:=n downto 1 do begin
      sr := 0.0;
      for k := pred(i) downto 1 do if
      (xpair[j,colver[k]] > sr)
        then sr := xpair[j,colver[k]];
      for k := pred(i) downto 1 do if (j = colver[k])
             then sr := 1.0;
      if ((sr < minprob)) then begin
        minprob := sr;
        colver[i] := j;
       end;
     end; (* for j *)
(* now colver[i] has the smallest maximum xpair with colver[1]..colver[i-1] *)
    end; (* for i *)
   threshold[i] := minprob; // 1234
   lastthreshold := minprob;
  end; (* not identified *)

  for i:=1 to n do ar[rank[i]] := i;
  writeln(outf,
   'A new ranking of the vertices was determined to bring out the block structure.');
  writeln(outf,'New ranking is ');
  case n of
    1..99: begin
             for i:=1 to n do write(outf,i:2,' '); writeln(outf);
             for i:=1 to n do write(outf,ar[i]:2,' '); writeln(outf);
           end;
    100..999: begin
             for i:=1 to n do begin
               write(outf,i:3,' ');
               if ((i mod 100) = 0) then writeln(outf);
              end;
             writeln(outf);
             if (i > 100) then writeln(outf);
             for i:=1 to n do begin
               write(outf,ar[i]:3,' ');
               if ((i mod 100) = 0) then writeln(outf);
              end;
             writeln(outf);
             if (i > 100) then writeln(outf);
             end;
    else    begin
             for i:=1 to n do begin
               write(outf,i:4,' ');
               if ((i mod 100) = 0) then writeln(outf);
              end;
             writeln(outf);
             if (i > 100) then writeln(outf);
             for i:=1 to n do begin
               write(outf,ar[i]:4,' ');
               if ((i mod 100) = 0) then writeln(outf);
              end;
             writeln(outf);
             if (i > 100) then writeln(outf);
           end;
    end; (* case *)
  writeln(outf);
  width0 := width;
  if (r0 + 2*r1 <= 9) then width := 1 else width := 2;

//  if (n <= 80) then begin
    writeln(outf,'Recoded observed adjacency matrix in new ranking is as follows: ');
    Head(1,'');
    for i:=1 to n do begin
      write(outf,ar[i]:4,'  ');
      for j:=1 to n do begin
        if (i <> j) then write(outf,y[ar[i],ar[j]]:width) else
          for h:=1 to width do write(outf,' ');
       end;
      writeln(outf);
     end; (* i *)
    writeln(outf);
//    end;     (* 80 *)

  width := width0;
(* Output of main results *)

  if (n <= 80) then
    begin

  writeln(outf,'The following matrices give the first digit after the decimal point (rounded).');
  if (n <= 100) then writeln(outf,'Two digits are given on file ',outfn1,'.');
  if (n <= 100) then writeln(outf1,'The matrices give the first two digits after the decimal point (rounded).');
  Heading(outf,4,'Matrix of pairwise color equality');
  if (n <= 100) then Heading(outf1,4,'Matrix of pairwise color equality');
  if (n <= 100) then Head(2,'') else Head(1,'');
  for i:=1 to n do begin
    write(outf,ar[i]:4,'  ');
    if (n <= 100) then write(outf1,ar[i]:4,'  ');
    for j:=1 to n do begin
      if (i <> j) then writed(xpair[ar[i],ar[j]],2)
       else begin for h:=1 to width do write(outf,' ');
            if (n <= 100) then write(outf1,'   '); end;
      if (((j mod 100) = 1) and (j > 1)and (n <= 100)) then
        begin writeln(outf1); write(outf1,'         '); end;
     end;
    writeln(outf); if (n <= 100) then writeln(outf1);
   end; (* i *)
   end; (* 80 *)

  sx := 2.0*sx/(n*(n-1.0));
  sx1 := 2.0*sx1/(n*(n-1.0));
  writeln(outf);
  writeln(outf,'Average of pairwise = ',sx1:6:4,' .');
  writeln(outf,'H_x = Average of 4*pairwise*(1-pairwise) = ',sx:6:4,' .');
  writeln(lf,'                           H_x = ',sx:6:4);
  writeln(outf1,'                           H_x = ',sx:6:4);
  write(outf,'(Given the first average, H_x must be ');
  writeln(outf,'between 0 and ',4*sx1*(1.0-sx1):6:4,').');
  writeln(outf);
  if (not identified) then begin
    writeln(outf,'Best separated vertices for identifying the ',c,' classes:');
    for i:=1 to c do begin
      write(outf,colver[i]:4);
      if (i >= 2) then writeln(outf,'  (errors ',threshold[i]:7:4,')') else writeln(outf);
     end; (* i *)
    writeln(outf);
   end; (* if *)

  if identified then begin
    Heading(outf,4,'Matrix of color probabilities of vertices');
    writeln(outf,'   ');for j:=1 to c do write(outf,j:6); writeln(outf);
    for i:=1 to n do begin
      write(outf,ar[i]:4,'  ');
      for j:=1 to c do write(outf,classprob[ar[i],j]:6:3);
      writeln(outf);
     end; (* i *)
    writeln(outf);
    writeln(outf,'Averages of posterior predictive distribution of number of vertices');
    writeln(outf,'  with the same color as this vertex ');
    writeln(outf,'vertex  av. number');
    for i:=1 to n do writeln(outf,ar[i]:6,'   ',predm[ar[i]]:6:2);
    writeln(outf);

    Heading(outf,4,'Posterior means of probabilities of colors');
    if (n <= 100) then Heading(outf1,4,'Standard deviations of probabilities of colors');
    for i:=1 to c do writeln(outf,i:2,thetax[i]:8:3,'  (',theta2[i]:6:3,')');
    writeln(outf);

    for k:=1 to r do begin
      writeln(outf,'Probabilities of tie value '+st(k)+' = '+symbol(k)+' between colors.');
      write(outf,' '); for j:=1 to c do write(outf,j:6); writeln(outf);
      for i:=1 to c do begin
        write(outf,i:3);
        for j:=1 to c do write(outf,fetac(k,i,j):6:3);
        writeln(outf);
       end; (* i *)
      writeln(outf);
      if (calcsd and (n <= 100)) then begin
        writeln(outf1);
        writeln(outf1,'Posterior standard deviations for probabilities tie value '+symbol(k));
        write(outf1,' '); for i:=1 to c do write(outf1,i:6); writeln(outf1);
        for i:=1 to c do begin
          write(outf1,i:3);
          for j:=1 to c do write(outf1,fetac2(k,i,j):6:3);
          writeln(outf1);
         end; (* i *)
        writeln(outf1);
       end; (* if calcsd *)
      end; (* k *)
   end else (* not identified *) begin
    Heading(outf1,4,'Class probabilities of vertices theta(x(i))');
    for i:=1 to n do writeln(outf1,ar[i]:4,thetax[ar[i]]:9:4);
    writeln(outf1);
    writeln(outf1,'Averages of posterior predictive distribution of number of vertices');
    writeln(outf1,'  with the same color as this vertex ');
    for i:=1 to n do writeln(outf1,ar[i]:4,' ',predm[ar[i]]:6:2);
    writeln(outf1);

    if (n <= 80) then
      begin

    Heading(outf,4,'Probabilities of tie values between vertices');
    if (n <= 100) then Heading(outf1,4,'Posterior probabilities (and standard deviations) of relations between vertices');
    for k:=1 to r do begin
      Head(2,'Probabilities of tie value '+st(k)+' = '+symbol(k)+' between vertices');
      for i:=1 to n do begin
        write(outf,ar[i]:4,'  ');
        if (n <= 100) then write(outf1,ar[i]:4,'  ');
        for j:=1 to n do begin
          if (i <> j) then writed(fetax(k,ar[i],ar[j]),2)
            else begin
              for h:=1 to width do write(outf,' ');
              if (n <= 100) then write(outf1,'   '); end;
      if (((j mod 100) = 1) and (j > 1) and (n <= 100)) then
        begin writeln(outf1); write(outf1,'  '); end;
         end; (* j *)
        writeln(outf); if (n <= 100) then writeln(outf1);
       end; (* i *)
      if (calcsd and (n <= 100)) then begin
        writeln(outf1); writeln(outf1,'Posterior standard deviations ');
        for i:=1 to n do begin
          write(outf1,ar[i]:4,'  ');
          for j:=1 to n do begin
            if (i <> j) then writed(fetax2(k,ar[i],ar[j]),1)
              else write(outf1,'   ');
      if (((j mod 100) = 1) and (j > 1)) then
        begin writeln(outf1); write(outf1,'  '); end;
           end;
          writeln(outf1);
         end; (* i *)
        end; (* if calcsd *)
      end; (* k *)

      end; (* 80 *)

     StrictBlock;
    end; (* not identified *)
  Finalize(col);
  Finalize(rank);
  Finalize(ar);
//  Finalize(chosen);
  Finalize(last);
end; (* GibbsResults *)

procedure MultiGibbs.FinishGibbs(var outf, outf1: textfile);
begin
  if escape then begin
    writeln(outf);
    writeln(outf,'The user asked for an early stop of the program.');
   end;
end; (* FinishGibbs *)

procedure MultiGibbs.stopsthis(var outf, outf1: textfile);
begin
  closefile(outf); closefile(outf1);
end; (* stopsthis *)

begin (* unit MulBlo *)
  ntiny := 0;
end.
