
Unit RanGen;

Interface

uses b_cons;

type realvec = array[0..cmax] of real;
     Boolmat = array of array of Boolean;
     dvec = array[1..cmax] of real;

(* Random number generators come from Press, Teukolsky & Vetterling *)

FUNCTION random0: Extended;
(* shuffled random number generator
   on the basis of internal random number generator *)

FUNCTION random1: Extended;
(* nicest but a bit slow *)

FUNCTION random3: Extended;
(* Subtractive method *)

FUNCTION rando(version: integer): Extended;
(* randomx with x = version for version = 0, 1 or 3. *)

PROCEDURE init0(seed: longint);
(* Initialises random0 *)

PROCEDURE init3(seed: longint);
(* Initialises random3 *)

PROCEDURE init(version: integer; seed: longint);
(* Initialises randomx, with x = version *)

FUNCTION takeseed(version: integer): longint;

FUNCTION randomd(version, k: integer): integer;
(* random between 0 and k-1; uses randomx, where x = version *)

function ranp(version: integer; var p: realvec): integer;
(* draws at random between 0 and n according to probability vector p *)

FUNCTION gamma(w: Extended):Extended;
(* Gamma distribution according to Ahrens & Dieter *)

FUNCTION beta1(b1,b2: Extended):Extended;
(* Beta distribution according to Ahrens & Dieter *)

FUNCTION beta2(b1,b2: Extended):Extended;
(* Beta distribution according to Narayanan & Johnk *)

Procedure Dirichlet(p: integer; w: dvec; var x: dvec);
(* Dirichlet distribution *)

FUNCTION normrand: Extended;
(* Normal distribution according to Box-Muller; uses random3 *)

FUNCTION binomial(pp: real; n: integer): integer;
(* Binomial distribution; uses random3 *)

procedure BernGraph(p: extended; g: integer; var Mat: Boolmat);
(* Bernoulli graph, upper triangle only; uses random3 *)

procedure FixGraph(k: word; g: integer; var Mat: Boolmat);
(* Random graph with k edges, upper triangle only; uses random3 *)

Implementation

uses WBlib;

VAR
   glinext,glinextp: longint;
   glma: ARRAY [1..55] OF Extended;
   gly, m3seed, nru: extended;
   glv: ARRAY [1..97] OF extended;
   glix1,glix2,glix3: longint;
   gliy: integer;
   glir: ARRAY [1..97] OF longint;
   ridum0, ridum1, ridum3: longInt;
(* ridumx just checks whether the random number generator
   has been initialised or not;
   the other 3 variables are a memory for the reshuffling process *)
   nrb: boolean; (* nrb, nru used in NormRand *)

FUNCTION gammln(xx: real): real;
CONST
   stp = 2.50662827465;
   half = 0.5;
   one = 1.0;
   fpf = 5.5;
VAR
   x,tmp,ser: extended;
   j: integer;
   cof: ARRAY [1..6] OF extended;
BEGIN
   cof[1] := 76.18009173;
   cof[2] := -86.50532033;
   cof[3] := 24.01409822;
   cof[4] := -1.231739516;
   cof[5] := 0.120858003e-2;
   cof[6] := -0.536382e-5;
   x := xx-one;
   tmp := x+fpf;
   tmp := (x+half)*ln(tmp)-tmp;
   ser := one;
   FOR j := 1 to 6 DO BEGIN
      x := x+one;
      ser := ser+cof[j]/x
   END;
   gammln := (tmp+ln(stp*ser))
END;

FUNCTION ran0(VAR idum: longint): extended;
(* RAN0 assumes that the system random number generator is called RANDOM,
with no arguments. Programs using RAN0 must declare the following
variables
VAR
   gly: real;
   glv: ARRAY [1..97] OF real;
in the main routine. *)
VAR
   dum: real;
   j: integer;
BEGIN
   IF (idum < 0) THEN BEGIN
      idum := 1;
      FOR j := 1 to 97 DO dum := random;
      FOR j := 1 to 97 DO glv[j] := random;
      gly := random
   END;
   j := 1+trunc(97.0*gly);
   IF ((j > 97) OR (j < 1)) THEN BEGIN
      writeln(lf, 'pause in routine RAN0'); readln END;
   gly := glv[j];
   ran0 := gly;
   glv[j] := random
END;

FUNCTION ran1(VAR idum: longint): extended;
(* Programs using RAN1 must declare the following variables
VAR
   glix1,glix2,glix3: integer;
   glv: ARRAY [1..97] OF real;
in the main program. *)
CONST
   m1=259200;
   ia1=7141;
   ic1=54773;
   rm1=3.8580247e-6;   (* 1.0/m1 *)
   m2=134456;
   ia2=8121;
   ic2=28411;
   rm2=7.4373773e-6;   (* 1.0/m2 *)
   m3=243000;
   ia3=4561;
   ic3=51349;
VAR
   j: integer;
BEGIN
   IF  (idum < 0)  THEN BEGIN
      glix1 := (ic1-idum) MOD m1;
      glix1 := (ia1*glix1+ic1) MOD m1;
      glix2 := glix1 MOD m2;
      glix1 := (ia1*glix1+ic1) MOD m1;
      glix3 := glix1 MOD m3;
      FOR j := 1 to 97 DO BEGIN
         glix1 := (ia1*glix1+ic1) MOD m1;
         glix2 := (ia2*glix2+ic2) MOD m2;
         glv[j] := (glix1+glix2*rm2)*rm1
      END;
      idum := 1
   END;
   glix1 := (ia1*glix1+ic1) MOD m1;
   glix2 := (ia2*glix2+ic2) MOD m2;
   glix3 := (ia3*glix3+ic3) MOD m3;
   j := 1 + (97*glix3) DIV m3;
   IF ((j > 97) OR (j < 1)) THEN BEGIN
      writeln(lf, 'pause in routine RAN1'); readln
   END;
   ran1 := glv[j];
   glv[j] := (glix1+glix2*rm2)*rm1
END;

FUNCTION ran3(VAR idum: longint): Extended;
CONST
   mbig=4.0e6;
   mz=0.0;
   fac=2.5e-7; (* 1/mbig *)
VAR
   i,ii,k: integer;
   mj,mk: Extended;
BEGIN
   IF (idum < 0) THEN BEGIN
      mj := m3seed+idum;
      (* The following IF block is mj := mj MOD mbig; for Extended variables. *)
      IF mj>=0.0 THEN mj := mj-mbig*trunc(mj/mbig)
         ELSE mj := mbig-abs(mj)+mbig*trunc(abs(mj)/mbig);
      glma[55] := mj;
      mk := 1;
      FOR i := 1 to 54 DO BEGIN
         ii := 21*i MOD 55;
         glma[ii] := mk;
         mk := mj-mk;
         IF (mk < mz) THEN mk := mk+mbig;
         mj := glma[ii]
      END;
      FOR k := 1 to 4 DO BEGIN
         FOR i := 1 to 55 DO BEGIN
            glma[i] := glma[i]-glma[1+((i+30) MOD 55)];
            IF (glma[i] < mz) THEN glma[i] := glma[i]+mbig
         END
      END;
      glinext := 0;
      glinextp := 31;
      idum := 1
   END; (* if idum < 0 *)
   glinext := glinext+1;
   IF (glinext = 56) THEN glinext := 1;
   glinextp := glinextp+1;
   IF (glinextp = 56) THEN glinextp := 1;
   mj := glma[glinext]-glma[glinextp];
   IF (mj < mz) THEN mj := mj+mbig;
   glma[glinext] := mj;
   ran3 := 0.00000000001 + 0.99999999998*mj*fac
END; (* ran3 *)

FUNCTION gamdev(ia: integer; VAR idum: longint): Extended;
(* gamma variable for integer shape parameter *)
VAR
   am,e,s,v1,v2,x,y,h: Extended;
   j: integer;
BEGIN
   IF (ia < 1) THEN BEGIN
      writeln(lf, 'pause in routine GAMDEV'); readln END;
   IF (ia < 6) THEN BEGIN
      x := 1.0;
      FOR j := 1 to ia DO BEGIN h:=ran3(idum);
                          IF h>1.e-12 THEN x := x*h
                                 ELSE x :=x*1.e-12;
                          END;
      x := -ln(x);
   END ELSE BEGIN
      REPEAT
         REPEAT
            REPEAT
               v1 := 2.0*ran3(idum)-1.0; IF v1<1.e-12 THEN v1:=1.e-12;
               v2 := 2.0*ran3(idum)-1.0;
            UNTIL ((sqr(v1)+sqr(v2)) <= 1.0);
            y := v2/v1;
            am := ia-1;
            s := sqrt(2.0*am+1.0);
            x := s*y+am;
         UNTIL (x > 1.e-12);
         e := (1.0+sqr(y))*exp(am*ln(x/am)-s*y);
      UNTIL (ran3(idum) <= e)
   END;
   gamdev := x
END; (* gamdev *)

FUNCTION gammaran01(w: Extended; VAR idum: longint):Extended;
LABEL 1;
VAR u1,u2,x0,x1,x2,k,x: Extended;
BEGIN
IF (w<=0) THEN
  BEGIN writeln(lf, 'pause in routine GAMMARAN01');
      writeln(lf, 'shape parameter ',w); readln 
  END;
1:    u1:=ran3(idum); u2:=ran3(idum);
      k:=1+w*exp(-1);
      x0:=k*u1;
      IF (x0>1) THEN BEGIN x2:=-ln((k-x0)/w);
                           IF (u2<=exp((w-1)*ln(x2))) THEN x:=x2
                                                      ELSE GOTO 1;
                     END
             ELSE BEGIN (* if u1 = 0 so that x0 = 0 then things go wrong *)
                           IF (x0<=1.e-12) THEN x0:=1.e-12;
                           x1:=exp(ln(x0)/w);
                           IF u2<=exp(-x1) THEN x:=x1
                                           ELSE GOTO 1;
                     END;
      gammaran01:=x;
END; (* gammaran01 *)

FUNCTION gammaran1(w: Extended; VAR idum:longint):Extended;
VAR u,x,y: Extended;
BEGIN IF (w<=0) THEN
  BEGIN writeln(lf, 'pause in routine GAMMARAN1'); 
      writeln(lf, 'shape parameter ',w); readln 
  END;
      REPEAT BEGIN u:=ran3(idum);
                   y:=-ln(ran3(idum));
             END
      UNTIL (ln(u) <= (w-1)*(ln(y)+1-y));
      x:=w*y;
      gammaran1:=x;
END; (* gammaran1 *)

FUNCTION gammaran(w: Extended; VAR idum:longint):Extended;
BEGIN IF (w<=0) THEN
  BEGIN writeln(lf, 'pause in routine GAMMARAN');
        writeln(lf, 'shape parameter ',w); readln
  END;
      IF (w<1) THEN gammaran:=gammaran01(w,idum)
          ELSE    gammaran:=gammaran1(w,idum);
END; (* gammaran *)

FUNCTION betaran1(VAR b1,b2: Extended; VAR idum: longint):Extended;
VAR y1,y2: Extended;
BEGIN if ((b1 <= 0) or (b2 <= 0)) then begin
       writeln; writeln(lf, 'betaran1 called with b1 = ',b1:12:4,', b2 = ',b2:12:4);
      end;
      y1:=gammaran(b1,idum);
      y2:=gammaran(b2,idum);
      betaran1:=y1/(y1+y2);
END;

FUNCTION betaran2(VAR b1,b2: Extended; VAR idum: longint):Extended;
VAR u1,u2,y1,y2: Extended;
BEGIN REPEAT BEGIN u1:=ran3(idum); u2:=ran3(idum);
                   IF u1<=0 then u1:=0.0001;
                   IF u2<=0 then u2:=0.0001;
                   y1:=exp(ln(u1)/b1);
                   y2:=exp(ln(u2)/b2);
             END
      UNTIL (y1+y2 <=1);
      betaran2:=y1/(y1+y2);
END;

FUNCTION NORMRAND: extended;
(* Normal distribution according to Box-Muller procedure *)
    VAR U1,U2,S: extended;
    i: integer;
BEGIN
  if nrb then normrand := nru else begin
    REPEAT
      U1:=2.0*RANDOM3-1.0;
      U2:=2.0*RANDOM3-1.0;
      S:=SQR(U1)+SQR(U2);
     UNTIL S<1.0;
    if (S < 1e-40) then begin
      normrand := 0.0; nru := 0.0;
     end else begin
      NORMRAND:=U1*SQRT(-2.0*LN(S)/S);
      nru:=U2*SQRT(-2.0*LN(S)/S);
     end;
   end; (* if *)
  nrb := not nrb;

{  s := 0; for i:=1 to 12 do s := s + random3;
  normrand := s-6; }
END; (* normrand *)

PROCEDURE init0(seed: longint);
(* Initialises random0 *)
begin
 ridum0 := -1;
 randseed := seed;
end;

PROCEDURE init3(seed: longint);
(* Initialises random3 *)
begin
 ridum3 := -1;
 m3seed := seed;
 nrb := false;
end;

PROCEDURE init(version: integer; seed: longint);
var a: char;
begin
  case version of
    0: init0(seed);
    1: begin
        writeln(lf); writeln(lf, 'Initialisation of random number generator,');
        writeln(lf, 'version 1, has not been implemented.');
        writeln(lf, 'Use another version if you want to initialise.');
//        write(' ... press a key ... '); a := readkey;
        Halt;
       end;
    3: init3(seed);
   end; (* case *)
end;

FUNCTION takeseed(version: integer): longint;
var s: longint;
begin
  case version of
    0: takeseed := randseed;
    3: begin
         s := trunc(4.0e6*random3);
         init3(s);
         takeseed := s;
        end;
   end; (* case *)
end;

FUNCTION random3: Extended;
begin
random3 := ran3(ridum3);
end;

FUNCTION random0: Extended;
(* shuffled random number generator
   on the basis of internal random number generator *)
begin
random0 := ran0(ridum0);
end;

FUNCTION random1: Extended;
begin
random1 := ran1(ridum1);
end;

FUNCTION rando(version: integer): Extended;
var a: Char;
begin
  case version of
     0: rando := random0;
     1: rando := random1;
     3: rando := random3;
     else begin
 writeln(lf, 'Call of rando in unit RanGen with non-implemented version.');
     readln(a);
     Halt;  end;
   end; (* case *)
end; (* rando *)


FUNCTION randomd(version, k: integer): integer;
var r: integer;
begin
  case version of
    0: r := trunc(random0*k);
    1: r := trunc(random1*k);
    3: r := trunc(random3*k);
    else begin
 writeln(lf, 'Call of randomd in unit RanGen with non-implemented version.');
     Halt; end;
   end; (* case *)
  if (r >= k) then randomd := k-1 else randomd := r;
end;

function ranp(version: integer; var p: realvec): integer;
(* draws at random between 0 and n according to probability vector p *)
var i: integer; s,r: real;
begin
  r := rando(version);
  s := 0;
  i:= -1;
  if (r >= 1.0) then r := 0.9999999999;
  repeat
    Inc(i);
    s := s+p[i];
   until (s >= r);
  ranp := i;
end;

FUNCTION gamma(w: Extended):Extended;
begin
gamma := gammaran(w,ridum3);
end;

FUNCTION beta1(b1,b2: Extended):Extended;
begin
beta1 := betaran1(b1,b2,ridum3);
end;

FUNCTION beta2(b1,b2: Extended):Extended;
begin
beta2 := betaran2(b1,b2,ridum3);
end;

Procedure Dirichlet(p: integer; w: dvec; var x: dvec);
(* Dirichlet distribution *)
const tiny = 1e-30;
var i: integer;
 xx: dvec;
 s: real;
 c: char;
begin
  if (p <= 0) or (p > cmax) then begin
    writeln;
    writeln(lf, 'Dirichlet called for the impossible value p = ',p);
    writeln(lf, 'Program stops.');
    writeln(lf, '   ... press a key to continue .... ');
    Halt;
   end;
  s := 0.0;
  for i:=p downto 1 do begin
    xx[i] := gamma(w[i]);
    s := s+xx[i];
   end; (* i *)
  if (s > tiny) then   (* s <= tiny has a negligible probability *)
    for i:=p downto 1 do x[i] := xx[i]/s
   else for i:=p downto 1 do x[i] := 1.0/p;
end;

FUNCTION binomial(pp: real; n: integer): integer;
LABEL 1;
VAR
   am,em,en,g,angle: real;
   oldg,p,pc: real;
   pclog,plog,pold,sq,t,y: real;
   j,nold,bnl: integer;
BEGIN
   nold := -1;
   pold := -1.0;
   IF (pp <= 0.5) THEN p := pp ELSE p := 1.0-pp;
   am := n*p;
   IF (n < 25) THEN BEGIN
      bnl := 0;
      FOR j := 1 to n DO BEGIN
         IF (random3 < p) THEN inc(bnl)
      END
   END ELSE IF (am < 1.0) THEN BEGIN
      g := exp(-am);
      t := 1.0;
      FOR j := 0 to n DO BEGIN
         t := t*random3;
         IF (t < g) THEN GOTO 1
      END;
      j := n;
1:      bnl := j
   END ELSE BEGIN
      IF (n <> nold) THEN BEGIN
         en := n;
         oldg := gammln(en+1.0);
         nold := n
      END;
      IF (p <> pold) THEN BEGIN
         pc := 1.0-p;
         plog := ln(p);
         pclog := ln(pc);
         pold := p
      END;
      sq := sqrt(2.0*am*pc);
      REPEAT
         REPEAT
            angle := pi*random3;
            y := sin(angle)/cos(angle);
            em := sq*y+am;
         UNTIL ((em >= 0.0) AND (em < en+1.0));
         em := trunc(em);
         t := 1.2*sq*(1.0+sqr(y))*exp(oldg-gammln(em+1.0)
            -gammln(en-em+1.0)+em*plog+(en-em)*pclog);
      UNTIL (random3 <= t);
      bnl := round(em)
   END;
   IF (pp > 0.5) THEN bnl := n-bnl;
   binomial := bnl
END;

procedure BernGraph(p: extended; g: integer; var Mat: Boolmat);
var i,j: integer; a : Char;
begin
if ((p < 0.0) or (p > 1.0)) then begin
  writeln;
  writeln(lf, 'Procedure BernGraph of RanGen called with wrong parameter.');
  writeln(lf, 'p = ',p);
  writeln(lf, 'press a key to continue ... ');
 end;
if (g > High(Mat)) then begin
  writeln;
  writeln(lf, 'Procedure BernGraph of RanGen called with wrong parameter.');
  writeln(lf, 'g = ',g);
  writeln(lf, 'press a key to continue ... ');
 end;
for j:=g downto 1 do for i:= j-1 downto 1 do
  Mat[i,j] := (random3 < p);
end;

procedure FixGraph(k: word; g: integer; var Mat: Boolmat);
var i,j: integer; kk, ss: word; a: Char;
begin
if ((k < 0) or (k > (g*(g-1)) div 2)) then begin
  writeln;
  writeln(lf, 'Procedure FixGraph of RanGen called with wrong parameter.');
  writeln(lf, 'k = ',k, ' , g = ',g);
  writeln(lf, 'press a key to continue ... ');
 end;
if (g > High(mat)) then begin
  writeln;
  writeln(lf, 'Procedure FixGraph of RanGen called with wrong parameter.');
  writeln(lf, 'g = ',g);
  writeln(lf, 'press a key to continue ... ');
 end;
kk := k; ss := g; ss := (ss*(g-1)) div 2 ;
for j:=g downto 1 do for i:=j-1 downto 1 do
  if (kk > 0) then begin
   if (ss*random3 < kk) then begin
     Mat[i,j] := true; Inc(kk,-1); Inc(ss,-1); end
    else begin
     Mat[i,j] := false; Inc(ss,-1); end;
    end else Mat[i,j] := false;
if ( kk > 0) then begin
  writeln;
  writeln(lf, 'Wrong parameters ',k:6,g:6,' in FixGraph (unit RanGen).');
  Halt;
 end;
end;


begin (* unit *)
ridum0 := -1;
ridum1 := -1;
ridum3 := -1;
m3seed := 1618033.0;
nrb := false;
end.